# test vector utilities

import math as maths

import numpy as np
from numpy.testing import assert_array_almost_equal

import resqpy.olio.vector_utilities as vec


def test_simple_functions():
    assert_array_almost_equal(vec.zero_vector(), (0.0, 0.0, 0.0))
    assert_array_almost_equal(vec.add((1.2, 3.4), (5.6, 7.8)), np.array([6.8, 11.2]))
    assert_array_almost_equal(vec.subtract((5.6, 7.8, 10.0), (1.2, 3.4, -1.0)), np.array([4.4, 4.4, 11.0]))
    assert_array_almost_equal(vec.elemental_multiply((1.2, 3.4), (5.6, 7.8)), (6.72, 26.52))
    assert_array_almost_equal(vec.amplify((1.0, 2.0, 4.5), 2.5), (2.5, 5.0, 11.25))
    assert_array_almost_equal(vec.v_3d((23.4, -98.7)), (23.4, -98.7, 0.0))
    assert_array_almost_equal(vec.v_3d((23.4, 45.4, -98.7)), np.array((23.4, 45.4, -98.7)))


def test_unit_vectors():
    v_set = np.array([(3.0, 4.0, 0.0), (3.7, -3.7, 3.7), (0.0, 0.0, 1.0), (0.0, 0.0, 0.0)])
    one_over_root_three = 1.0 / maths.sqrt(3.0)
    expected = np.array([(3.0 / 5.0, 4.0 / 5.0, 0.0), (one_over_root_three, -one_over_root_three, one_over_root_three),
                         (0.0, 0.0, 1.0), (0.0, 0.0, 0.0)])
    for v, e in zip(v_set, expected):
        assert_array_almost_equal(vec.unit_vector(v), e)
    assert_array_almost_equal(vec.unit_vectors(v_set), expected)
    azi = [0.0, -90.0, 120.0, 180.0, 270.0, 360.0]
    expected = np.array([(0.0, 1.0, 0.0), (-1.0, 0.0, 0.0), (maths.cos(maths.pi / 6), -0.5, 0.0), (0.0, -1.0, 0.0),
                         (-1.0, 0.0, 0.0), (0.0, 1.0, 0.0)])
    for a, e in zip(azi, expected):
        assert_array_almost_equal(vec.unit_vector_from_azimuth(a), e)


def test_angles():
    pi_by_2 = maths.pi / 2.0
    assert maths.isclose(vec.radians_from_degrees(90.0), pi_by_2)
    assert_array_almost_equal(vec.radians_from_degrees((0.0, -90.0, 120.0, 270.0)),
                              (0.0, -pi_by_2, maths.pi * 2 / 3, pi_by_2 * 3))
    assert maths.isclose(vec.radians_from_degrees(vec.degrees_from_radians(1.2647)), 1.2647)
    assert_array_almost_equal(vec.degrees_from_radians((0.0, -pi_by_2, maths.pi * 2 / 3, pi_by_2 * 3)),
                              (0.0, -90.0, 120.0, 270.0))
    assert maths.isclose(vec.degrees_from_radians(maths.pi), 180.0)


def test_azimuth():
    vectors = np.array([
        (0.0, 1.0, 0.0),  # north
        (27.3, 0.0, 2341.0),  # east
        (0.0, -4.5e6, -12000.0),  # south
        (-0.001, 0.0, 23.2),  # west
        (0.0, 0.0, 100.0),  # no bearing (expected to return 0.0 for azimuth(), NaN for azimuths())
        (123.45, -123.45, 0.03),  # south east
        (maths.cos(maths.radians(30.0)), 0.5, 0.0)  # bearing of 60 degrees
    ])
    expected = np.array((0.0, 90.0, 180.0, 270.0, 0.0, 135.0, 60.0))
    for v, e in zip(vectors, expected):
        assert maths.isclose(vec.azimuth(v), e)
        assert maths.isclose(vec.azimuth(v[:2]), e)
    expected[4] = np.NaN
    assert_array_almost_equal(vec.azimuths(vectors), expected)
    assert_array_almost_equal(vec.azimuths(vectors[:, :2]), expected)


def test_inclination():
    vectors = np.array([(0.0, 0.0, 1.0), (0.0, 0.0, -458.21), (0.0, 233.67, 233.67), (910.0, 0.0, 910.0),
                        (0.0, 0.0, 0.0)])
    expected = (0.0, 180.0, 45.0, 45.0, 90.0)  # last is an arbitrary value returned by the function for zero vector
    for v, e in zip(vectors, expected):
        assert maths.isclose(vec.inclination(v), e)
    assert maths.isclose(vec.inclination((456.7, -456.7, 456.7)),
                         90.0 - vec.degrees_from_radians(maths.acos(maths.sqrt(2) / maths.sqrt(3))))


def test_clockwise():
    p1 = (4.4, 4.4)
    p2 = (6.8, 6.8)
    p3 = (-21.0, -21.0)
    assert maths.isclose(vec.clockwise(p1, p2, p3), 0.0)
    p1 = [1.0, 1.0, -45.0]
    p2 = [2.0, 8.0, 23.1]
    p3 = [5.0, 3.0, -11.0]
    assert vec.clockwise(p1, p2, p3) > 0.0
    p1, p2 = np.array(p2), np.array(p1)
    p3 = np.array(p3)
    assert vec.clockwise(p1, p2, p3) < 0.0


def test_clockwise_triangles():
    p = np.array([(1.0, 1.0, 1.0), (2.0, 2.0, -1.0), (1.0, 2.0, 0.0), (2.0, 1.0, 0.0)])
    t = np.array([(0, 1, 2), (0, 1, 3)], dtype = int)
    e_xy = np.array([-1.0, 1.0])
    assert np.all(vec.clockwise_triangles(p, t) * e_xy > 0.0)
    e_xz = np.array([1.0, -1.0])
    assert np.all(vec.clockwise_triangles(p, t, projection = 'xz') * e_xz > 0.0)
    e_yz = np.array([-1.0, 1.0])
    assert np.all(vec.clockwise_triangles(p, t, projection = 'yz') * e_yz > 0.0)


def test_points_in_triangles():
    p = np.array([(1.0, 1.0, 1.0), (2.0, 2.0, -1.0), (1.0, 2.0, 0.0), (2.0, 1.0, 0.0)])
    t = np.array([(0, 1, 2), (0, 1, 3)], dtype = int)
    d = np.array([(0.5, 1.5, 0.0), (1.25, 1.75, 0.0), (1.75, 1.25, -3.0), (1.5, 1.5, 1.5), (1.0, 1.0, 3.0),
                  (1.5, 0.5, 1.0), (1.5, 2.5, 0.0), (0.5, 0.5, 0.5), (2.5, 1.5, -1.0)])
    e_no_edge = np.zeros((len(t), len(d)), dtype = bool)
    e_no_edge[0, 1] = True
    e_no_edge[1, 2] = True
    e_edge = e_no_edge.copy()
    e_edge[:, 3] = True
    e_edge[:, 4] = True
    r_no_edge = vec.points_in_triangles(p, t, d, projection = 'xy', edged = False)
    r_edge = vec.points_in_triangles(p, t, d, projection = 'xy', edged = True)
    assert np.all(r_no_edge == e_no_edge)
    assert np.all(r_edge == e_edge)



from resqpy import model


def test_empty_model():
    _ = model.Model()
    return


def test_all_imports():
    from resqpy import (crs, derived_model, fault, grid, grid_surface, lines, organize, property, rq_import, surface,
                        time_series, well)

    # The line below prevents IDEs from deleting the above
    _ = (crs, derived_model, fault, grid, grid_surface, lines, organize, property, rq_import, surface, time_series,
         well)

    #    from resqpy.olio import *
    return



import numpy as np
import pandas as pd

import resqpy.well
from resqpy.model import Model


def test_WellboreFrame(example_model_and_crs):
    # Test that all attributes are correctly saved and loaded from disk

    # --------- Arrange ----------
    # Create a WellboreFrame object in memory
    # Load example model from a fixture
    model, crs = example_model_and_crs
    epc_path = model.epc_file

    # Create a trajectory
    well_name = 'Banoffee'
    elevation = 100
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    mds = np.array([300, 310, 330])
    zs = mds - elevation
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [1, 2, 3],
        'Y': [1, 2, 3],
        'Z': zs,
    })
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        well_name = well_name,
                                        md_datum = datum,
                                        length_uom = 'm')
    trajectory.write_hdf5()
    trajectory.create_xml()

    # Create a wellbore frame object
    wellbore_frame_mds = np.array([305, 315])
    wellbore_frame = resqpy.well.WellboreFrame(parent_model = model,
                                               trajectory = trajectory,
                                               mds = wellbore_frame_mds,
                                               title = 'WellboreFrame_1',
                                               originator = 'Person_1')

    # ----------- Act ---------
    # Save to disk
    wellbore_frame.write_hdf5()
    wellbore_frame.create_xml()
    wellbore_frame_uuid = wellbore_frame.uuid
    model.store_epc()
    model.h5_release()

    # Clear memory
    del model, datum, trajectory, wellbore_frame

    # Reload from disk
    model2 = Model(epc_file = epc_path)
    wellbore_frame_2 = resqpy.well.WellboreFrame(parent_model = model2, uuid = wellbore_frame_uuid)

    # ----------- Assert ---------
    assert wellbore_frame_2.node_count == 2
    np.testing.assert_equal(wellbore_frame_2.node_mds, wellbore_frame_mds)


def test_create_feature_and_intrepretation(example_model_and_crs):
    # Test that WellboreFeature and WellboreInterpretation objects can be added to WellboreFrame object

    # --------- Arrange ----------
    # Create a WellboreFrame object without a well feature
    # Load example model from a fixture
    model, crs = example_model_and_crs
    epc_path = model.epc_file

    # Create a trajectory
    well_name = 'Banoffee'
    elevation = 100
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    mds = np.array([300, 310, 330])
    zs = mds - elevation
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [1, 2, 3],
        'Y': [1, 2, 3],
        'Z': zs,
    })
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        well_name = well_name,
                                        md_datum = datum,
                                        length_uom = 'm')
    trajectory.write_hdf5()
    trajectory.create_xml()

    # Create a wellbore frame object
    wellbore_frame_mds = np.array([305, 315])
    wellbore_frame = resqpy.well.WellboreFrame(parent_model = model,
                                               trajectory = trajectory,
                                               mds = wellbore_frame_mds,
                                               title = 'WellboreFrame_1',
                                               originator = 'Person_1')

    assert wellbore_frame.wellbore_feature is None
    assert wellbore_frame.wellbore_interpretation is None

    # ----------- Act ---------
    wellbore_frame.create_feature_and_interpretation()

    # ----------- Assert ---------
    assert wellbore_frame.wellbore_feature is not None
    assert wellbore_frame.wellbore_feature.feature_name == wellbore_frame.title



import numpy as np
import pandas as pd
from numpy.testing import assert_array_almost_equal
from pandas.testing import assert_frame_equal


def test_trajectory_iterators(example_model_with_well):
    # Test that trajectories can be found via iterators

    model, well_interp, datum, traj = example_model_with_well

    # Iterate via wells
    uuids_1 = []
    for well in model.iter_wellbore_interpretations():
        for traj in well.iter_trajectories():
            uuids_1.append(traj.uuid)

    # Iterate directly
    uuids_2 = []
    for traj in model.iter_trajectories():
        uuids_2.append(traj.uuid)

    # Two sets of trajectories should match, assuming all trajectories have a parent well
    assert len(uuids_1) > 0
    assert sorted(uuids_1) == sorted(uuids_2)


def test_logs(example_model_with_logs):

    model, well_interp, datum, traj, frame, log_collection = example_model_with_logs

    # Check logs can be extracted from resqml dataset

    # Check exactly one wellbore frame exists in the model
    discovered_frames = 0
    for trajectory in model.iter_trajectories():
        for frame in trajectory.iter_wellbore_frames():
            discovered_frames += 1
    assert discovered_frames == 1

    # Check MDs
    mds = frame.node_mds
    assert isinstance(mds, np.ndarray)
    assert_array_almost_equal(mds, [1, 2, 3, 4])

    # Check logs
    log_list = list(log_collection.iter_logs())
    assert len(log_list) == 2

    # TODO: would be nice to write: log_collection.get_curve("GR")
    gr = log_list[0]
    assert gr.title == "GR"
    assert gr.uom == "gAPI"
    assert_array_almost_equal(gr.values(), [1, 2, 1, 2])

    nphi = log_list[1]
    assert nphi.title == 'NPHI'

    # TODO: get more units working
    # assert nphi.uom == "v/v"
    assert_array_almost_equal(nphi.values(), [0.1, 0.1, np.NaN, np.NaN])


def test_logs_conversion(example_model_with_logs):

    model, well_interp, datum, traj, frame, log_collection = example_model_with_logs

    # Pandas
    df = log_collection.to_df()
    df_expected = pd.DataFrame(data = {"GR": [1, 2, 1, 2], "NPHI": [0.1, 0.1, np.NaN, np.NaN]}, index = [1, 2, 3, 4])
    assert_frame_equal(df_expected, df, check_dtype = False)

    # LAS
    las = log_collection.to_las()
    assert las.well.WELL.value == 'well A'

    gr = las.get_curve("GR")
    assert gr.unit.casefold() == "GAPI".casefold()
    assert_array_almost_equal(gr.data, [1, 2, 1, 2])

    nphi = las.get_curve("NPHI")
    # assert nphi.unit == "GAPI"
    assert_array_almost_equal(nphi.data, [0.1, 0.1, np.NaN, np.NaN])



"""_deviation_survey.py: resqpy well module providing deviation survey class"""

# todo: create a trajectory from a deviation survey, assuming minimum curvature

version = '18th November 2021'

# Nexus is a registered trademark of the Halliburton Company
# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import pandas as pd

import resqpy.olio.uuid as bu
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
import resqpy.weights_and_measures as bwam
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns

from .well_utils import _as_optional_array, extract_xyz, load_hdf5_array
from ._md_datum import MdDatum


class DeviationSurvey(BaseResqpy):
    """Class for RESQML wellbore deviation survey.

    RESQML documentation:

       Specifies the station data from a deviation survey.

       The deviation survey does not provide a complete specification of the
       geometry of a wellbore trajectory. Although a minimum-curvature
       algorithm is used in most cases, the implementation varies sufficiently
       that no single algorithmic specification is available as a data transfer
       standard.

       Instead, the geometry of a RESQML wellbore trajectory is represented by
       a parametric line, parameterized by the MD.

       CRS and units of measure do not need to be consistent with the CRS and
       units of measure for wellbore trajectory representation.
    """

    resqml_type = 'DeviationSurveyRepresentation'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 title = None,
                 deviation_survey_root = None,
                 represented_interp = None,
                 md_datum = None,
                 md_uom = 'm',
                 angle_uom = 'dega',
                 measured_depths = None,
                 azimuths = None,
                 inclinations = None,
                 station_count = None,
                 first_station = None,
                 is_final = False,
                 originator = None,
                 extra_metadata = None):
        """Load or create a DeviationSurvey object.

        If uuid is given, loads from XML. Else, create new. If loading from disk, other
        parameters will be overwritten.

        Args:
           parent_model (model.Model): the model which the new survey belongs to
           uuid (uuid.UUID): If given, loads from disk. Else, creates new.
           title (str): Citation title
           deviation_survey_root: DEPCRECATED. If given, load from disk.
           represented_interp (wellbore interpretation): if present, is noted as the wellbore
              interpretation object which this deviation survey relates to
           md_datum (MdDatum): the datum that the depths for this survey are measured from
           md_uom (string, default 'm'): a resqml length unit of measure applicable to the
              measured depths; should be 'm' or 'ft'
           angle_uom (string): a resqml angle unit; should be 'dega' or 'rad'
           measured_depths (np.array): 1d array
           azimuths (np.array): 1d array
           inclinations (np.array): 1d array
           station_count (int): length of measured_depths, azimuths & inclinations
           first_station (tuple): (x, y, z) of first point in survey, in crs for md datum
           is_final (bool): whether survey is a finalised deviation survey
           originator (str): name of author
           extra_metadata (dict, optional): extra metadata key, value pairs

        Returns:
           DeviationSurvey

        Notes:
           this method does not create an xml node, nor write hdf5 arrays
        """

        self.is_final = is_final
        self.md_uom = bwam.rq_length_unit(md_uom)

        self.angles_in_degrees = angle_uom.strip().lower().startswith('deg')
        """boolean: True for degrees, False for radians (nothing else supported). Should be 'dega' or 'rad'"""

        # Array data
        self.measured_depths = _as_optional_array(measured_depths)
        self.azimuths = _as_optional_array(azimuths)
        self.inclinations = _as_optional_array(inclinations)

        if station_count is None and measured_depths is not None:
            station_count = len(measured_depths)
        self.station_count = station_count
        self.first_station = first_station

        # Referenced objects
        self.md_datum = md_datum  # md datum is an object in its own right, with a related crs!
        self.wellbore_interpretation = represented_interp

        # TODO: remove deviation_survey_root, use just uuid

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata,
                         root_node = deviation_survey_root)

    @classmethod
    def from_data_frame(cls,
                        parent_model,
                        data_frame,
                        md_datum = None,
                        md_col = 'MD',
                        azimuth_col = 'AZIM_GN',
                        inclination_col = 'INCL',
                        x_col = 'X',
                        y_col = 'Y',
                        z_col = 'Z',
                        md_uom = 'm',
                        angle_uom = 'dega'):
        """Load MD, aximuth & inclination data from a pandas data frame.

        Args:
           parent_model (model.Model): the parent resqml model
           data_frame: a pandas dataframe holding the deviation survey data
           md_datum (MdDatum object): the datum that the depths for this survey are measured from
           md_col (string, default 'MD'): the name of the column holding measured depth values
           azimuth_col (string, default 'AZIM_GN'): the name of the column holding azimuth values relative
              to the north direction (+ve y axis) of the coordinate reference system
           inclination_col (string, default 'INCL'): the name of the column holding inclination values
           x_col (string, default 'X'): the name of the column holding an x value in the first row
           y_col (string, default 'Y'): the name of the column holding an Y value in the first row
           z_col (string, default 'Z'): the name of the column holding an z value in the first row
           md_uom (string, default 'm'): a resqml length unit of measure applicable to the
              measured depths; should be 'm' or 'ft'
           angle_uom (string, default 'dega'): a resqml angle unit of measure applicable to both
              the azimuth and inclination data

        Returns:
           DeviationSurvey

        Note:
           The X, Y & Z columns are only used to set the first station location (from the first row)
        """

        for col in [md_col, azimuth_col, inclination_col, x_col, y_col, z_col]:
            assert col in data_frame.columns
        station_count = len(data_frame)
        assert station_count >= 2  # vertical well could be hamdled by allowing a single station in survey?
        #         self.md_uom = bwam.p_length_unit(md_uom)

        start = data_frame.iloc[0]

        return cls(parent_model = parent_model,
                   station_count = station_count,
                   md_datum = md_datum,
                   md_uom = md_uom,
                   angle_uom = angle_uom,
                   first_station = (start[x_col], start[y_col], start[z_col]),
                   measured_depths = data_frame[md_col].values,
                   azimuths = data_frame[azimuth_col].values,
                   inclinations = data_frame[inclination_col].values,
                   is_final = True)  # assume this is a finalised deviation survey

    @classmethod
    def from_ascii_file(cls,
                        parent_model,
                        deviation_survey_file,
                        comment_character = '#',
                        space_separated_instead_of_csv = False,
                        md_col = 'MD',
                        azimuth_col = 'AZIM_GN',
                        inclination_col = 'INCL',
                        x_col = 'X',
                        y_col = 'Y',
                        z_col = 'Z',
                        md_uom = 'm',
                        angle_uom = 'dega',
                        md_datum = None):
        """Load MD, aximuth & inclination data from an ascii deviation survey file.

        Arguments:
           parent_model (model.Model): the parent resqml model
           deviation_survey_file (string): the filename of an ascii file holding the deviation survey data
           comment_character (string): the character to be treated as introducing comments
           space_separated_instead_of_csv (boolea, default False): if False, csv format expected;
              if True, columns are expected to be seperated by white space
           md_col (string, default 'MD'): the name of the column holding measured depth values
           azimuth_col (string, default 'AZIM_GN'): the name of the column holding azimuth values relative
              to the north direction (+ve y axis) of the coordinate reference system
           inclination_col (string, default 'INCL'): the name of the column holding inclination values
           x_col (string, default 'X'): the name of the column holding an x value in the first row
           y_col (string, default 'Y'): the name of the column holding an Y value in the first row
           z_col (string, default 'Z'): the name of the column holding an z value in the first row
           md_uom (string, default 'm'): a resqml length unit of measure applicable to the
              measured depths; should be 'm' or 'ft'
           angle_uom (string, default 'dega'): a resqml angle unit of measure applicable to both
              the azimuth and inclination data
           md_datum (MdDatum object): the datum that the depths for this survey are measured from

        Returns:
           DeviationSurvey

        Note:
           The X, Y & Z columns are only used to set the first station location (from the first row)
        """

        try:
            df = pd.read_csv(deviation_survey_file,
                             comment = comment_character,
                             delim_whitespace = space_separated_instead_of_csv)
            if df is None:
                raise Exception
        except Exception:
            log.error('failed to read ascii deviation survey file ' + deviation_survey_file)
            raise

        return cls.from_data_frame(parent_model,
                                   df,
                                   md_col = md_col,
                                   azimuth_col = azimuth_col,
                                   inclination_col = inclination_col,
                                   x_col = x_col,
                                   y_col = y_col,
                                   z_col = z_col,
                                   md_uom = md_uom,
                                   angle_uom = angle_uom,
                                   md_datum = md_datum)

    def _load_from_xml(self):
        """Load attributes from xml and associated hdf5 data.

        This is invoked as part of the init method when an existing uuid is given.

        Returns:
           [bool]: True if successful
        """

        # Get node from self.uuid
        node = self.root
        assert node is not None

        # Load XML data
        self.md_uom = rqet.length_units_from_node(rqet.find_tag(node, 'MdUom', must_exist = True))
        self.angle_uom = rqet.find_tag_text(node, 'AngleUom', must_exist = True)
        self.station_count = rqet.find_tag_int(node, 'StationCount', must_exist = True)
        self.first_station = extract_xyz(rqet.find_tag(node, 'FirstStationLocation', must_exist = True))
        self.is_final = rqet.find_tag_bool(node, 'IsFinal')

        # Load HDF5 data
        mds_node = rqet.find_tag(node, 'Mds', must_exist = True)
        load_hdf5_array(self, mds_node, 'measured_depths')
        azimuths_node = rqet.find_tag(node, 'Azimuths', must_exist = True)
        load_hdf5_array(self, azimuths_node, 'azimuths')
        inclinations_node = rqet.find_tag(node, 'Inclinations', must_exist = True)
        load_hdf5_array(self, inclinations_node, 'inclinations')

        # Set related objects
        self.md_datum = self._load_related_datum()
        self.represented_interp = self._load_related_wellbore_interp()

        # Validate
        assert self.measured_depths is not None
        assert len(self.measured_depths) > 0

        return True

    def create_xml(self,
                   ext_uuid = None,
                   md_datum_root = None,
                   md_datum_xyz = None,
                   add_as_part = True,
                   add_relationships = True,
                   title = None,
                   originator = None):
        """Creates a deviation survey representation xml element from this DeviationSurvey object.

        arguments:
           ext_uuid (uuid.UUID): the uuid of the hdf5 external part holding the deviation survey arrays
           md_datum_root: the root xml node for the measured depth datum that the deviation survey depths
              are based on
           md_datum_xyz: TODO: document this
           add_as_part (boolean, default True): if True, the newly created xml node is added as a part
              in the model
           add_relationships (boolean, default True): if True, a relationship xml part is created relating the
              new deviation survey part to the measured depth datum part
           title (string): used as the citation Title text; should usually refer to the well name in a
              human readable way
           originator (string, optional): the name of the human being who created the deviation survey part;
              default is to use the login name

        returns:
           the newly created deviation survey xml node
        """

        assert self.station_count > 0

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()

        md_datum_root = self.__get_md_datum_root(md_datum_root = md_datum_root, md_datum_xyz = md_datum_xyz)

        # Create root node, write citation block
        ds_node = super().create_xml(title = title, originator = originator, add_as_part = False)

        mds_values_node, azimuths_values_node, inclinations_values_node = self.__add_sub_elements_to_root_node(
            ds_node = ds_node)

        self.model.create_md_datum_reference(md_datum_root, root = ds_node)
        self.model.create_solitary_point3d('FirstStationLocation', ds_node, self.first_station)
        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'Mds', root = mds_values_node)
        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'Azimuths', root = azimuths_values_node)
        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'Inclinations', root = inclinations_values_node)
        interp_root = self.__create_wellbore_interpretation_ref_node(ds_node = ds_node)
        self.__add_as_part_and_add_relationships(ds_node = ds_node,
                                                 md_datum_root = md_datum_root,
                                                 interp_root = interp_root,
                                                 ext_uuid = ext_uuid,
                                                 add_as_part = add_as_part,
                                                 add_relationships = add_relationships)

        return ds_node

    def __get_md_datum_root(self, md_datum_root, md_datum_xyz):
        """ Ensures that the root node for the MdDatum object that the DeviationSurvey depths are based on exists.

        If not, a root node will be created and returned.
        """

        if md_datum_root is None:
            if self.md_datum is None:
                if md_datum_xyz is None:
                    raise ValueError("Must provide a MD Datum for the DeviationSurvey")
                self.md_datum = MdDatum(self.model, location = md_datum_xyz)
            if self.md_datum.root is None:
                md_datum_root = self.md_datum.create_xml()
            else:
                md_datum_root = self.md_datum.root
        assert md_datum_root is not None
        return md_datum_root

    def __add_sub_elements_to_root_node(self, ds_node):
        """Appends sub-elements to the DeviationSurvey object's root node."""

        if_node = rqet.SubElement(ds_node, ns['resqml2'] + 'IsFinal')
        if_node.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
        if_node.text = str(self.is_final).lower()

        sc_node = rqet.SubElement(ds_node, ns['resqml2'] + 'StationCount')
        sc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        sc_node.text = str(self.station_count)

        md_uom = rqet.SubElement(ds_node, ns['resqml2'] + 'MdUom')
        md_uom.set(ns['xsi'] + 'type', ns['eml'] + 'LengthUom')
        md_uom.text = bwam.rq_length_unit(self.md_uom)

        angle_uom = rqet.SubElement(ds_node, ns['resqml2'] + 'AngleUom')
        angle_uom.set(ns['xsi'] + 'type', ns['eml'] + 'PlaneAngleUom')
        if self.angles_in_degrees:
            angle_uom.text = 'dega'
        else:
            angle_uom.text = 'rad'

        mds = rqet.SubElement(ds_node, ns['resqml2'] + 'Mds')
        mds.set(ns['xsi'] + 'type', ns['resqml2'] + 'DoubleHdf5Array')
        mds.text = rqet.null_xml_text

        mds_values_node = rqet.SubElement(mds, ns['resqml2'] + 'Values')
        mds_values_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Hdf5Dataset')
        mds_values_node.text = rqet.null_xml_text

        azimuths = rqet.SubElement(ds_node, ns['resqml2'] + 'Azimuths')
        azimuths.set(ns['xsi'] + 'type', ns['resqml2'] + 'DoubleHdf5Array')
        azimuths.text = rqet.null_xml_text

        azimuths_values_node = rqet.SubElement(azimuths, ns['resqml2'] + 'Values')
        azimuths_values_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Hdf5Dataset')
        azimuths_values_node.text = rqet.null_xml_text

        inclinations = rqet.SubElement(ds_node, ns['resqml2'] + 'Inclinations')
        inclinations.set(ns['xsi'] + 'type', ns['resqml2'] + 'DoubleHdf5Array')
        inclinations.text = rqet.null_xml_text

        inclinations_values_node = rqet.SubElement(inclinations, ns['resqml2'] + 'Values')
        inclinations_values_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Hdf5Dataset')
        inclinations_values_node.text = rqet.null_xml_text

        return mds_values_node, azimuths_values_node, inclinations_values_node

    def __create_wellbore_interpretation_ref_node(self, ds_node):
        """Create a reference node for the WellboreInterpretation object and add to the DeviationSurvey root node. """

        interp_root = None
        if self.wellbore_interpretation is not None:
            interp_root = self.wellbore_interpretation.root
            self.model.create_ref_node('RepresentedInterpretation',
                                       rqet.find_nested_tags_text(interp_root, ['Citation', 'Title']),
                                       bu.uuid_from_string(interp_root.attrib['uuid']),
                                       content_type = 'obj_WellboreInterpretation',
                                       root = ds_node)
        return interp_root

    def __add_as_part_and_add_relationships(self, ds_node, md_datum_root, interp_root, ext_uuid, add_as_part,
                                            add_relationships):
        """Add the newly created DeviationSurvey object's root node as a part in the model and add reciprocal

        relationships.
        """

        if add_as_part:
            self.model.add_part('obj_DeviationSurveyRepresentation', self.uuid, ds_node)
            if add_relationships:
                # todo: check following relationship
                self.model.create_reciprocal_relationship(ds_node, 'destinationObject', md_datum_root, 'sourceObject')
                if interp_root is not None:
                    self.model.create_reciprocal_relationship(ds_node, 'destinationObject', interp_root, 'sourceObject')
                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(ds_node, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')

    def write_hdf5(self, file_name = None, mode = 'a'):
        """Create or append to an hdf5 file, writing datasets for the measured depths, azimuths, and inclinations."""

        # NB: array data must all have been set up prior to calling this function
        h5_reg = rwh5.H5Register(self.model)
        h5_reg.register_dataset(self.uuid, 'Mds', self.measured_depths, dtype = float)
        h5_reg.register_dataset(self.uuid, 'Azimuths', self.azimuths, dtype = float)
        h5_reg.register_dataset(self.uuid, 'Inclinations', self.inclinations, dtype = float)
        h5_reg.write(file = file_name, mode = mode)

    def _load_related_datum(self):
        """Return related MdDatum object from XML if present."""

        md_datum_uuid = bu.uuid_from_string(rqet.find_tag(rqet.find_tag(self.root, 'MdDatum'), 'UUID'))
        if md_datum_uuid is not None:
            md_datum_part = 'obj_MdDatum_' + str(md_datum_uuid) + '.xml'
            md_datum = MdDatum(self.model, md_datum_root = self.model.root_for_part(md_datum_part, is_rels = False))
        else:
            md_datum = None
        return md_datum

    def _load_related_wellbore_interp(self):
        """Return related wellbore interp object from XML if present."""

        interp_uuid = rqet.find_nested_tags_text(self.root, ['RepresentedInterpretation', 'UUID'])
        if interp_uuid is None:
            represented_interp = None
        else:
            represented_interp = rqo.WellboreInterpretation(self.model, uuid = interp_uuid)
        return represented_interp



import os
import pandas as pd
import numpy as np
import pytest

from resqpy.well.well_object_funcs import add_wells_from_ascii_file, well_name, add_blocked_wells_from_wellspec, add_logs_from_cellio, lookup_from_cellio
import resqpy.well
from resqpy.grid import RegularGrid


def test_add_wells_from_ascii_file(example_model_and_crs, caplog):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    crs_uuid = crs.uuid
    ascii_file = os.path.join(model.epc_directory, 'ascii.dat')
    well_name = 'Well_1'

    source_df = pd.DataFrame([[well_name, 200, 1, 1, 150], [well_name, 250, 2, 2, 200], [well_name, 300, 2, 3, 250]],
                             columns = ['NotaWell', 'MD', 'X', 'Y', 'Z'])
    with open(ascii_file, 'w') as fp:
        for col in source_df.columns:
            fp.write(f' {col}')
        fp.write('\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                if col == 'NotaWell':
                    fp.write(f' {(row[col])}')
                else:
                    fp.write(f' {row[col]}')
            fp.write('\n')

    # Use incorrect name for the well column
    # --------- Act ----------
    with pytest.raises(AssertionError):
        (feature_list, interpretation_list, trajectory_list,
         md_datum_list) = add_wells_from_ascii_file(model = model,
                                                    crs_uuid = crs_uuid,
                                                    trajectory_file = ascii_file,
                                                    well_col = None,
                                                    space_separated_instead_of_csv = True)

    # Use correct name for well column
    # --------- Act ----------
    (feature_list2, interpretation_list2, trajectory_list2,
     md_datum_list2) = add_wells_from_ascii_file(model = model,
                                                 crs_uuid = crs_uuid,
                                                 trajectory_file = ascii_file,
                                                 well_col = 'NotaWell',
                                                 space_separated_instead_of_csv = True)

    test_trajectory = trajectory_list2[0]
    # --------- Assert ----------
    assert len(feature_list2) == len(interpretation_list2) == len(trajectory_list2) == len(md_datum_list2) == 1
    np.testing.assert_equal(test_trajectory.measured_depths, np.array([200., 250., 300.]))


def test_best_root_for_object(example_model_with_well):

    # --------- Arrange ----------
    model, well_interp, datum, traj = example_model_with_well
    traj_title = traj.title
    wellbore_feature_uuid = model.uuid_for_part(part_name = model.part(obj_type = 'obj_WellboreFeature'))
    wellbore_marker_frame = resqpy.well.WellboreMarkerFrame(parent_model = model, trajectory_uuid = traj.uuid)
    wellbore_marker_frame.create_xml()
    wellbore_frame = resqpy.well.WellboreFrame(parent_model = model, trajectory = traj)
    wellbore_frame.create_xml()
    deviation_survey = resqpy.well.DeviationSurvey(parent_model = model,
                                                   md_datum = datum,
                                                   measured_depths = [200, 210, 250],
                                                   azimuths = [10, 10, 10],
                                                   inclinations = [45, 45, 45],
                                                   first_station = (0, 0, 0))
    deviation_survey.create_xml()

    # --------- Act ----------
    citation_title1 = well_name(well_object = well_interp, model = model)
    citation_title2 = well_name(well_object = wellbore_marker_frame, model = model)
    citation_title3 = well_name(well_object = wellbore_frame, model = model)
    citation_title4 = well_name(well_object = datum, model = model)
    citation_title5 = well_name(well_object = deviation_survey, model = model)
    citation_title6 = well_name(well_object = wellbore_feature_uuid, model = model)

    # --------- Assert ----------
    assert citation_title1 == citation_title2 == citation_title3 == citation_title6 == traj_title
    assert citation_title4 == citation_title5 == datum.title


def test_add_blocked_wells_from_wellspec(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (3, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    wellspec_file = os.path.join(model.epc_directory, 'wellspec.dat')
    well_name = 'DOGLEG'
    source_df = pd.DataFrame([[2, 2, 1, 0.0, 0.0, 0.0, 0.25, 0.9], [2, 2, 2, 0.45, -90.0, 2.5, 0.25, 0.9],
                              [2, 3, 2, 0.45, -90.0, 1.0, 0.20, 0.9], [2, 3, 3, 0.0, 0.0, -0.5, 0.20, 0.9]],
                             columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW', 'PPERF'])
    with open(wellspec_file, 'w') as fp:
        fp.write(F'WELLSPEC {well_name}\n')
        for col in source_df.columns:
            fp.write(f' {col:>6s}')
        fp.write('\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                if col in ['IW', 'JW', 'L']:
                    fp.write(f' {int(row[col]):6d}')
                else:
                    fp.write(f' {row[col]:6.2f}')
            fp.write('\n')

    # --------- Act ----------
    count = add_blocked_wells_from_wellspec(model = model, grid = grid, wellspec_file = wellspec_file)

    # --------- Assert ----------
    assert count == 1


def test_add_logs_from_cellio_file(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (5, 4, 4),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)

    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    elevation = 100
    # Create a measured depth datum
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    mds = np.array([100, 210, 230, 240, 250])
    zs = mds - elevation
    well_name = 'Coconut'
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [25, 50, 75, 100, 100],
        'Y': [25, -50, -75, -100, -100],
        'Z': zs,
        'WELL': ['Coconut', 'Coconut', 'Coconut', 'Coconut', 'Coconut']
    })

    # Create a trajectory from dataframe
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        well_name = well_name,
                                        md_datum = datum,
                                        length_uom = 'm')
    trajectory.write_hdf5()
    trajectory.create_xml()

    # Create a blocked well using the trajectory
    bw = resqpy.well.BlockedWell(model,
                                 well_name = well_name,
                                 grid = grid,
                                 trajectory = trajectory,
                                 use_face_centres = True)
    bw.cell_grid_link = bw.map_cell_and_grid_indices()  # TODO: confirm this is valid
    bw.write_hdf5()
    bw.create_xml()

    cellio_file = os.path.join(model.epc_directory, 'cellio.dat')
    source_df = pd.DataFrame(
        [[1, 1, 1, 25, 25, 0, 26, 26, 1, 120, 0.12], [2, 2, 1, 26, -26, 126, 27, -27, 127, 117, 0.20],
         [2, 3, 1, 27, -27, 127, 28, -28, 128, 135, 0.15]],
        columns = [
            'i_index unit1 scale1', 'j_index unit1 scale1', 'k_index unit1 scale1', 'x_in unit1 scale1',
            'y_in unit1 scale1', 'z_in unit1 scale1', 'x_out unit1 scale1', 'y_out unit1 scale1', 'z_out unit1 scale1',
            'Perm unit1 scale1', 'Poro unit1 scale1'
        ])

    with open(cellio_file, 'w') as fp:
        fp.write('1.0\n')
        fp.write('Undefined\n')
        fp.write(f'{well_name} terrible day\n')
        fp.write('11\n')
        for col in source_df.columns:
            fp.write(f' {col}\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                fp.write(f' {int(row[col])}')
            fp.write('\n')

    # --------- Act ----------
    add_logs_from_cellio(blockedwell = bw, cellio = cellio_file)

    # --------- Assert ----------
    assert 'cellio.Coconut.Perm' in model.titles()
    assert 'cellio.Coconut.Poro' in model.titles()


def test_lookup_from_cellio(example_model_and_crs):
    # --------- Arrange ----------
    model, crs = example_model_and_crs
    line = 'Perm continuous 23 34 45 67 87 99'

    # --------- Act ----------
    uuid = lookup_from_cellio(line = line, model = model)
    uuid2 = lookup_from_cellio(line = line, model = model)

    # --------- Assert ----------
    assert uuid is not None
    assert 'Perm' in model.titles()
    assert uuid2 is uuid



# test module for resqpy.olio.transmission.py

import math as maths
import os

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.transmission as rqtr
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec


def test_regular_grid_half_cell_transmission(tmp_path):

    def try_one_half_t_regular(model,
                               extent_kji = (2, 2, 2),
                               dxyz = (1.0, 1.0, 1.0),
                               perm_kji = (1.0, 1.0, 1.0),
                               ntg = 1.0,
                               darcy_constant = 1.0,
                               rotate = None,
                               dip = None):
        ones = np.ones(extent_kji)
        grid = grr.RegularGrid(model, extent_kji = extent_kji, dxyz = dxyz)
        if dip is not None:  # dip positive x axis downwards
            r_matrix = vec.rotation_matrix_3d_axial(1, dip)
            p = grid.points_ref(masked = False)
            p[:] = vec.rotate_array(r_matrix, p)
        if rotate is not None:  # rotate anticlockwise in xy plane (viewed from above)
            r_matrix = vec.rotation_matrix_3d_axial(2, rotate)
            p = grid.points_ref(masked = False)
            p[:] = vec.rotate_array(r_matrix, p)
        half_t = rqtr.half_cell_t(grid,
                                  perm_k = perm_kji[0] * ones,
                                  perm_j = perm_kji[1] * ones,
                                  perm_i = perm_kji[2] * ones,
                                  ntg = ntg * ones,
                                  darcy_constant = darcy_constant)
        expected = 2.0 * darcy_constant * np.array(
            (perm_kji[0] * dxyz[0] * dxyz[1] / dxyz[2], ntg * perm_kji[1] * dxyz[0] * dxyz[2] / dxyz[1],
             ntg * perm_kji[2] * dxyz[1] * dxyz[2] / dxyz[0]))
        assert np.all(np.isclose(half_t, expected.reshape(1, 1, 1, 3)))

    temp_epc = str(os.path.join(tmp_path, f"{bu.new_uuid()}.epc"))
    model = rq.Model(temp_epc, new_epc = True, create_basics = True)

    try_one_half_t_regular(model)
    try_one_half_t_regular(model, extent_kji = (3, 4, 5))
    try_one_half_t_regular(model, dxyz = (127.53, 21.05, 12.6452))
    try_one_half_t_regular(model, perm_kji = (123.23, 512.4, 314.7))
    try_one_half_t_regular(model, ntg = 0.7)
    try_one_half_t_regular(model, darcy_constant = 0.001127)
    try_one_half_t_regular(model,
                           extent_kji = (5, 4, 3),
                           dxyz = (84.23, 77.31, 15.823),
                           perm_kji = (0.6732, 298.14, 384.2),
                           ntg = 0.32,
                           darcy_constant = 0.008527)
    try_one_half_t_regular(model, rotate = 67.8)

    # should not have written anything, but try clean-up just in case
    try:
        os.remove(temp_epc)
    except Exception:
        pass
    try:
        os.remove(temp_epc[-4] + '.h5')
    except Exception:
        pass


def test_half_cell_t_2d_triangular_precursor_equilateral():
    side = 123.45
    height = side * maths.sin(vec.radians_from_degrees(60.0))
    # create a pair of equilateral triangles
    p = np.array([(0.0, 0.0), (side, 0.0), (side / 2.0, height), (side * 3.0 / 2.0, height)])
    t = np.array([(0, 1, 2), (1, 2, 3)], dtype = int)
    a, b = rqtr.half_cell_t_2d_triangular_precursor(p, t)
    a_over_l = a / b
    expected = side / (height / 3)
    assert a_over_l.shape == (2, 3)
    assert_array_almost_equal(a_over_l, expected)



import resqpy.rq_import as rqi
import resqpy.model as rq
import os

import numpy as np
from numpy.testing import assert_array_almost_equal

import pytest


## TODO: move to a general utilities area once complete
def simple_grid_corns(k_gap = False, righthanded = True, undefined = False):
    """Returns corner points for simple 2x2x2 grid"""
    origin_cell = np.array([
        [
            [
                [0, 2, 0],  # top, back, left, K+, J+, I-
                [1, 2, 0]
            ],  # top, back, right, K+, J+, I+
            [
                [0, 1, 0],  # top, front, left, K+, J-, I-
                [1, 1, 0]
            ]
        ],  # top, front, right, K+, J-, I+
        [
            [
                [0, 2, 1],  # bottom, back, left, K-, J+, I-
                [1, 2, 1]
            ],  # bottom, back, right, K-, J+, I+
            [
                [0, 1, 1],  # bottom, front, left, K-, J-, I-
                [1, 1, 1]
            ]
        ]
    ])  # bottom, front, right, K-, J-, I+

    if not righthanded:
        origin_cell[:, :, :, 1] = origin_cell[:, :, :, 1] * -1
        origin_cell[:, :, :, 1] += 2

    c000 = origin_cell.copy()
    c100 = origin_cell.copy()
    c010 = origin_cell.copy()
    c110 = origin_cell.copy()
    c001 = origin_cell.copy()
    c101 = origin_cell.copy()
    c011 = origin_cell.copy()
    c111 = origin_cell.copy()

    c100[:, :, :, 0] += 1,  # k0, j0, i1
    c110[:, :, :, 0] += 1  # k0, j1, i1
    c101[:, :, :, 0] += 1  # k1, j0, i1
    c111[:, :, :, 0] += 1  # k1, j1, i1

    if righthanded:
        c010[:, :, :, 1] -= 1  # k0, j1, i0
        c110[:, :, :, 1] -= 1  # k0, j1, i1
        c011[:, :, :, 1] -= 1  # k1, j1, i0
        c111[:, :, :, 1] -= 1  # k1, j1, i1
    else:
        c010[:, :, :, 1] += 1  # k0, j1, i0
        c110[:, :, :, 1] += 1  # k0, j1, i1
        c011[:, :, :, 1] += 1  # k1, j1, i0
        c111[:, :, :, 1] += 1  # k1, j1, i1

    c001[:, :, :, 2] += 1  # k1, j0, i0
    c101[:, :, :, 2] += 1  # k1, j0, i1
    c011[:, :, :, 2] += 1  # k1, j1, i0
    c111[:, :, :, 2] += 1  # k1, j1, i1

    corns = np.array(
        [
            [
                [
                    c000,  # k0, j0, i0
                    c100
                ],  # k0, j0, i1
                [
                    c010,  # k0, j1, i0
                    c110
                ]
            ],  # k0, j1, i1
            [
                [
                    c001,  # k1, j0, i0
                    c101
                ],  # k1, j0, i1
                [
                    c011,  # k1, j1, i0
                    c111
                ]
            ]
        ],
        dtype = 'float')  # k1, j1, i1

    if k_gap:
        corns[1, :, :, :, :, :, 2] += 1
    if undefined:
        corns[0, 0, 0, :, :, :, :] = np.nan
    return corns


def test_grid_from_cp_simple(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corns = simple_grid_corns()
    expected_points = np.array([[[[0, 2, 0], [1, 2, 0], [2, 2, 0]], [[0, 1, 0], [1, 1, 0], [2, 1, 0]],
                                 [[0, 0, 0], [1, 0, 0], [2, 0, 0]]],
                                [[[0, 2, 1], [1, 2, 1], [2, 2, 1]], [[0, 1, 1], [1, 1, 1], [2, 1, 1]],
                                 [[0, 0, 1], [1, 0, 1], [2, 0, 1]]],
                                [[[0, 2, 2], [1, 2, 2], [2, 2, 2]], [[0, 1, 2], [1, 1, 2], [2, 1, 2]],
                                 [[0, 0, 2], [1, 0, 2], [2, 0, 2]]]])

    # Act
    grid = rqi.grid_from_cp(model, cp_array = corns, crs_uuid = crs.uuid, ijk_handedness = None)

    # Assert
    assert grid is not None
    assert_array_almost_equal(grid.extent_kji, (2, 2, 2))
    assert grid.grid_is_right_handed
    assert grid.k_gaps is None
    assert grid.k_direction_is_down
    assert grid.pillar_shape == 'curved'
    assert grid.crs_uuid == crs.uuid
    assert grid.geometry_defined_for_all_cells_cached
    assert_array_almost_equal(grid.points_cached, expected_points)


def test_grid_from_cp_simple_inactive(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corns = simple_grid_corns()
    active_mask = np.ones((2, 2, 2))
    active_mask[1, 1, 1] = 0
    expected_inactive = ~active_mask.astype(bool)

    # Act
    grid = rqi.grid_from_cp(model,
                            cp_array = corns,
                            crs_uuid = crs.uuid,
                            ijk_handedness = None,
                            active_mask = active_mask)

    # Assert
    assert grid is not None
    assert_array_almost_equal(grid.extent_kji, (2, 2, 2))
    assert grid.grid_is_right_handed
    assert grid.k_gaps is None
    assert grid.k_direction_is_down
    assert grid.pillar_shape == 'curved'
    assert grid.crs_uuid == crs.uuid
    assert grid.geometry_defined_for_all_cells_cached
    assert grid.array_cell_geometry_is_defined is None
    assert_array_almost_equal(grid.inactive, expected_inactive)


def test_grid_from_cp_simple_nogeom(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corns = simple_grid_corns(undefined = True)
    expected_bool = np.ones((2, 2, 2))
    expected_bool[0, 0, 0] = 0

    # Act
    grid = rqi.grid_from_cp(model,
                            cp_array = corns,
                            crs_uuid = crs.uuid,
                            ijk_handedness = None,
                            geometry_defined_everywhere = False)

    # Assert
    assert grid is not None
    assert_array_almost_equal(grid.extent_kji, (2, 2, 2))
    assert grid.grid_is_right_handed
    assert grid.k_gaps is None
    assert grid.k_direction_is_down
    assert grid.pillar_shape == 'curved'
    assert grid.crs_uuid == crs.uuid
    assert not grid.geometry_defined_for_all_cells_cached
    assert_array_almost_equal(grid.array_cell_geometry_is_defined, expected_bool)


def test_grid_from_cp_simple_inactive_nogeom(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corns = simple_grid_corns(undefined = True)
    expected_bool = np.ones((2, 2, 2))
    expected_bool[0, 0, 0] = 0
    active_mask = np.ones((2, 2, 2))
    active_mask[1, 1, 1] = 0
    expected_active = active_mask.copy()
    expected_active[0, 0, 0] = 0
    expected_inactive = ~expected_active.astype(bool)

    # Act
    grid = rqi.grid_from_cp(model,
                            cp_array = corns,
                            crs_uuid = crs.uuid,
                            ijk_handedness = None,
                            geometry_defined_everywhere = False,
                            active_mask = active_mask)

    # Assert
    assert grid is not None
    assert_array_almost_equal(grid.extent_kji, (2, 2, 2))
    assert grid.grid_is_right_handed
    assert grid.k_gaps is None
    assert grid.k_direction_is_down
    assert grid.pillar_shape == 'curved'
    assert grid.crs_uuid == crs.uuid
    assert not grid.geometry_defined_for_all_cells_cached
    assert_array_almost_equal(grid.array_cell_geometry_is_defined, expected_active)
    assert_array_almost_equal(grid.inactive, expected_inactive)


def test_grid_from_cp_simple_left(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corns = simple_grid_corns(righthanded = False)

    # Act
    grid = rqi.grid_from_cp(model, cp_array = corns, crs_uuid = crs.uuid, ijk_handedness = None)

    # Assert
    assert grid is not None
    assert_array_almost_equal(grid.extent_kji, (2, 2, 2))
    assert not grid.grid_is_right_handed
    assert grid.k_gaps is None
    assert grid.k_direction_is_down
    assert grid.pillar_shape == 'curved'
    assert grid.crs_uuid == crs.uuid


def test_grid_from_cp_simple_straight(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corns = simple_grid_corns()

    # Act
    grid = rqi.grid_from_cp(model, cp_array = corns, crs_uuid = crs.uuid, known_to_be_straight = True)

    # Assert
    assert grid is not None
    assert_array_almost_equal(grid.extent_kji, (2, 2, 2))
    assert grid.grid_is_right_handed
    assert grid.k_gaps is None
    assert grid.k_direction_is_down
    assert grid.pillar_shape == 'straight'


def test_grid_from_cp_kgap(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corns = simple_grid_corns(k_gap = True)

    # Act
    grid = rqi.grid_from_cp(model, cp_array = corns, crs_uuid = crs.uuid)

    # Assert
    assert grid is not None
    assert_array_almost_equal(grid.extent_kji, (2, 2, 2))
    assert grid.grid_is_right_handed
    assert grid.k_gaps == 1
    assert grid.k_direction_is_down


def test_grid_from_cp_kgap_zvoid(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corns = simple_grid_corns()

    # Add a k-gap
    corns[1, :, :, :, :, :, 2] += 0.5

    # Act
    grid = rqi.grid_from_cp(model, cp_array = corns, crs_uuid = crs.uuid, max_z_void = 1)

    # Assert
    assert grid is not None
    assert_array_almost_equal(grid.extent_kji, (2, 2, 2))
    assert grid.grid_is_right_handed
    assert grid.k_gaps is None
    assert grid.k_direction_is_down


@pytest.mark.parametrize('surfaces,format,interp_and_feat,role,rqclass,newparts',
                         [(['Surface_zmap.dat'], 'zmap', False, 'map', 'surface', 2),
                          (['Surface_zmap.dat'], 'zmap', True, 'map', 'surface', 4),
                          (['Surface_roxartext.txt'], 'roxar', False, 'map', 'surface', 2),
                          (['Surface_roxartext.txt'], 'roxar', True, 'map', 'surface', 4),
                          (['Surface_roxartext.txt'], 'rms', False, 'map', 'surface', 2),
                          (['Surface_roxartext.txt'], 'rms', True, 'map', 'surface', 4),
                          (['Surface_tsurf.txt'], 'GOCAD-Tsurf', False, 'map', 'surface', 2),
                          (['Surface_tsurf.txt'], 'GOCAD-Tsurf', True, 'map', 'surface', 4),
                          (['Surface_zmap.dat', 'Surface_zmap.dat'], 'zmap', False, 'map', 'surface', 3),
                          (['Surface_zmap.dat', 'Surface_zmap.dat'], 'zmap', True, 'map', 'surface', 6),
                          (['Surface_zmap.dat'], 'zmap', False, 'pick', 'surface', 2),
                          (['Surface_zmap.dat'], 'zmap', False, 'pick', 'TriangulatedSet', 2)])
# (['Surface_zmap.dat'],'zmap',False,'pick','Grid2d',2)]) # TODO: Fails due to bug, Mesh.create_xml does not have an argument for crs_uuid
def test_add_surfaces(example_model_and_crs, test_data_path, surfaces, format, rqclass, interp_and_feat, role,
                      newparts):
    model, crs = example_model_and_crs
    model.store_epc()

    surface_paths = [os.path.join(test_data_path, surf) for surf in surfaces]

    rqi.add_surfaces(epc_file = model.epc_file,
                     surface_file_format = format,
                     surface_file_list = surface_paths,
                     surface_role = role,
                     rq_class = rqclass,
                     make_horizon_interpretations_and_features = interp_and_feat)

    model = rq.Model(model.epc_file)
    assert len(model.parts()) == newparts
    if rqclass in ['surface', 'TriangulatedSet']:
        assert len(model.parts_list_of_type('obj_TriangulatedSetRepresentation')) == len(surfaces)
    else:
        assert len(model.parts_list_of_type('obj_Grid2dRepresentation')) == len(surfaces)

    if interp_and_feat:
        assert len(model.parts_list_of_type('obj_HorizonInterpretation')) == len(surfaces)



import math as maths
import os

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

import resqpy.derived_model as rqdm
import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.property as rqp
import resqpy.time_series as rqts
import resqpy.weights_and_measures as bwam
import resqpy.surface as rqs
import resqpy.olio.xml_et as rqet
import resqpy.well as rqw
from resqpy.crs import Crs

from resqpy.property import property_kind_and_facet_from_keyword, guess_uom

# ---- Test PropertyCollection methods ---

# TODO


def test_property(tmp_path):
    epc = os.path.join(tmp_path, 'test_prop.epc')
    model = rq.new_model(epc)
    grid = grr.RegularGrid(model, extent_kji = (2, 3, 4))
    grid.write_hdf5()
    grid.create_xml()
    a1 = np.random.random(grid.extent_kji)
    p1 = rqp.Property.from_array(model,
                                 a1,
                                 source_info = 'random',
                                 keyword = 'NETGRS',
                                 support_uuid = grid.uuid,
                                 property_kind = 'net to gross ratio',
                                 indexable_element = 'cells',
                                 uom = 'm3/m3')
    a3 = np.random.random((grid.nk + 1, grid.nj + 1, grid.ni + 1))
    p3 = rqp.Property.from_array(model,
                                 a3,
                                 source_info = 'random',
                                 keyword = 'jiggle shared nodes',
                                 support_uuid = grid.uuid,
                                 property_kind = 'length',
                                 indexable_element = 'nodes',
                                 uom = 'm3')
    a4 = np.random.random((grid.nk, grid.nj, grid.ni, 2, 2, 2))
    p4 = rqp.Property.from_array(model,
                                 a4,
                                 source_info = 'random',
                                 keyword = 'jiggle nodes per cell',
                                 support_uuid = grid.uuid,
                                 property_kind = 'length',
                                 indexable_element = 'nodes per cell',
                                 uom = 'm3')
    pk = rqp.PropertyKind(model, title = 'facies', parent_property_kind = 'discrete')
    pk.create_xml()
    facies_dict = {0: 'background'}
    for i in range(1, 10):
        facies_dict[i] = 'facies ' + str(i)
    sl = rqp.StringLookup(model, int_to_str_dict = facies_dict, title = 'facies table')
    sl.create_xml()
    shape2 = tuple(list(grid.extent_kji) + [2])
    a2 = (np.random.random(shape2) * 10.0).astype(int)
    p2 = rqp.Property.from_array(model,
                                 a2,
                                 source_info = 'random',
                                 keyword = 'FACIES',
                                 support_uuid = grid.uuid,
                                 local_property_kind_uuid = pk.uuid,
                                 indexable_element = 'cells',
                                 discrete = True,
                                 null_value = 0,
                                 count = 2,
                                 string_lookup_uuid = sl.uuid)
    model.store_epc()
    model = rq.Model(epc)
    ntg_uuid = model.uuid(obj_type = p1.resqml_type, title = 'NETGRS')
    assert ntg_uuid is not None
    p1p = rqp.Property(model, uuid = ntg_uuid)
    assert np.all(p1p.array_ref() == p1.array_ref())
    assert p1p.uom() == 'm3/m3'
    facies_uuid = model.uuid(obj_type = p2.resqml_type, title = 'FACIES')
    assert facies_uuid is not None
    p2p = rqp.Property(model, uuid = facies_uuid)
    assert np.all(p2p.array_ref() == p2.array_ref())
    assert p2p.null_value() is not None and p2p.null_value() == 0
    grid = model.grid()
    jiggle_parts = model.parts(title = 'jiggle', title_mode = 'starts')
    assert len(jiggle_parts) == 2
    jiggle_shared_uuid = model.uuid(parts_list = jiggle_parts, title = 'shared', title_mode = 'contains')
    assert jiggle_shared_uuid is not None
    p3p = rqp.Property(model, uuid = jiggle_shared_uuid)
    assert p3p is not None
    assert p3p.array_ref().shape == (grid.nk + 1, grid.nj + 1, grid.ni + 1)
    jiggle_per_cell_uuid = model.uuid(parts_list = jiggle_parts, title = 'per cell', title_mode = 'ends')
    assert jiggle_per_cell_uuid is not None
    # four properties created here, plus 3 regular grid cell lengths properties
    assert grid.property_collection.number_of_parts() == 7
    collection = rqp.selective_version_of_collection(grid.property_collection,
                                                     property_kind = 'length',
                                                     uuid = jiggle_per_cell_uuid)
    assert collection is not None
    assert collection.number_of_parts() == 1
    p4p = rqp.Property.from_singleton_collection(collection)
    assert p4p is not None
    assert p4p.array_ref().shape == (grid.nk, grid.nj, grid.ni, 2, 2, 2)


def test_create_Property_from_singleton_collection(tmp_model):
    # Arrange
    grid = grr.RegularGrid(tmp_model, extent_kji = (2, 3, 4))
    grid.write_hdf5()
    grid.create_xml(add_cell_length_properties = True)
    collection = rqp.selective_version_of_collection(grid.property_collection,
                                                     property_kind = 'cell length',
                                                     facet_type = 'direction',
                                                     facet = 'J')
    assert collection.number_of_parts() == 1
    # Check property can be created
    prop = rqp.Property.from_singleton_collection(collection)
    assert np.all(np.isclose(prop.array_ref(), 1.0))


# ---- Test property kind parsing ---


def test_property_kinds():
    prop_kinds = bwam.valid_property_kinds()
    assert "angle per time" in prop_kinds
    assert "foobar" not in prop_kinds


# ---- Test infer_property_kind ---


@pytest.mark.parametrize(
    "input_name, input_unit, kind, facet_type, facet",
    [
        ('gamma ray API unit', 'gAPI', 'gamma ray API unit', None, None),  # Just a placeholder for now
        ('', '', 'Unknown', None, None),  # What should default return?
    ])
def test_infer_property_kind(input_name, input_unit, kind, facet_type, facet):
    kind_, facet_type_, facet_ = rqp.infer_property_kind(input_name, input_unit)
    assert kind == kind_
    assert facet_type == facet_type_
    assert facet == facet_


def test_bespoke_property_kind():
    model = rq.Model(create_basics = True)
    em = {'something': 'important', 'and_another_thing': 42}
    pk1 = rqp.PropertyKind(model, title = 'my kind of property', extra_metadata = em)
    pk1.create_xml()
    assert len(model.uuids(obj_type = 'PropertyKind')) == 1
    pk2 = rqp.PropertyKind(model, title = 'my kind of property', extra_metadata = em)
    pk2.create_xml()
    assert len(model.uuids(obj_type = 'PropertyKind')) == 1
    pk3 = rqp.PropertyKind(model, title = 'your kind of property', extra_metadata = em)
    pk3.create_xml()
    assert len(model.uuids(obj_type = 'PropertyKind')) == 2
    pk4 = rqp.PropertyKind(model, title = 'my kind of property')
    pk4.create_xml()
    assert len(model.uuids(obj_type = 'PropertyKind')) == 3
    pk5 = rqp.PropertyKind(model, title = 'my kind of property', parent_property_kind = 'discrete', extra_metadata = em)
    pk5.create_xml()
    assert len(model.uuids(obj_type = 'PropertyKind')) == 4
    pk6 = rqp.PropertyKind(model, title = 'my kind of property', parent_property_kind = 'continuous')
    pk6.create_xml()
    assert len(model.uuids(obj_type = 'PropertyKind')) == 4
    assert pk6 == pk4


def test_string_lookup():
    model = rq.Model(create_basics = True)
    d = {1: 'peaty', 2: 'muddy', 3: 'sandy', 4: 'granite'}
    sl = rqp.StringLookup(model, int_to_str_dict = d, title = 'stargazing')
    sl.create_xml()
    assert sl.length() == 5 if sl.stored_as_list else 4
    assert sl.get_string(3) == 'sandy'
    assert sl.get_index_for_string('muddy') == 2
    d2 = {0: 'nothing', 27: 'something', 173: 'amazing', 1072: 'brilliant'}
    sl2 = rqp.StringLookup(model, int_to_str_dict = d2, title = 'head in the clouds')
    assert not sl2.stored_as_list
    assert sl2.length() == 4
    assert sl2.get_string(1072) == 'brilliant'
    assert sl2.get_string(555) is None
    assert sl2.get_index_for_string('amazing') == 173
    sl2.create_xml()
    assert set(model.titles(obj_type = 'StringTableLookup')) == set(['stargazing', 'head in the clouds'])
    assert sl != sl2


def test_constant_array_expansion(tmp_path):
    epc = os.path.join(tmp_path, 'boring.epc')
    model = rq.new_model(epc)
    grid = grr.RegularGrid(model, extent_kji = (2, 3, 4))
    grid.write_hdf5()
    grid.create_xml()
    p1 = rqp.Property.from_array(model,
                                 cached_array = None,
                                 source_info = 'test',
                                 keyword = 'constant pi',
                                 support_uuid = grid.uuid,
                                 property_kind = 'continuous',
                                 indexable_element = 'cells',
                                 const_value = maths.pi,
                                 expand_const_arrays = True)
    p2 = rqp.Property.from_array(model,
                                 cached_array = None,
                                 source_info = 'test',
                                 keyword = 'constant three',
                                 support_uuid = grid.uuid,
                                 property_kind = 'discrete',
                                 discrete = True,
                                 indexable_element = 'cells',
                                 const_value = 3,
                                 expand_const_arrays = False)
    model.store_epc()
    # re-open the model and check that p1 appears as an ordinary array whilst p2 is still a constant
    model = rq.Model(epc)
    p1p = rqp.Property(model, uuid = p1.uuid)
    p2p = rqp.Property(model, uuid = p2.uuid)
    assert p1p.constant_value() is None
    assert p2p.constant_value() == 3
    assert_array_almost_equal(p1p.array_ref(), np.full(grid.extent_kji, maths.pi))
    assert np.all(p2p.array_ref() == 3) and p2p.array_ref().shape == tuple(grid.extent_kji)


def test_property_extra_metadata(tmp_path):
    # set up test model with a grid
    epc = os.path.join(tmp_path, 'em_test.epc')
    model = rq.new_model(epc)
    grid = grr.RegularGrid(model, extent_kji = (2, 3, 4))
    grid.write_hdf5()
    grid.create_xml()
    model.store_epc()
    # add a grid property with some extra metadata
    a = np.arange(24).astype(float).reshape((2, 3, 4))
    em = {'important': 'something', 'also': 'nothing'}
    uuid = rqdm.add_one_grid_property_array(epc, a, 'length', title = 'nonsense', uom = 'm', extra_metadata = em)
    # re-open the model and check that extra metadata has been preserved
    model = rq.Model(epc)
    p = rqp.Property(model, uuid = uuid)
    for item in em.items():
        assert item in p.extra_metadata.items()


def test_points_properties(tmp_path):
    epc = os.path.join(tmp_path, 'points_test.epc')
    model = rq.new_model(epc)

    extent_kji = (2, 3, 4)
    ensemble_size = 5
    faulted_ensemble_size = 2
    time_series_size = 6

    # create a geological time series
    years = [int(t) for t in np.linspace(-252170000, -66000000, num = time_series_size, dtype = int)]
    ts = rqts.GeologicTimeSeries.from_year_list(model, year_list = years, title = 'Mesozoic time series')
    ts.create_xml()
    ts_uuid = ts.uuid
    assert ts.timeframe == 'geologic'

    #  create a simple grid without an explicit geometry and ensure it has a property collection initialised
    grid = grr.RegularGrid(model,
                           extent_kji = extent_kji,
                           origin = (0.0, 0.0, 1000.0),
                           dxyz = (100.0, 100.0, -10.0),
                           title = 'unfaulted grid')
    # patch K direction, even though it won't be stored in xml
    grid.k_direction_is_down = False
    grid.grid_is_right_handed = not grid.grid_is_right_handed
    # generate xml and establish a property collection for the grid
    grid.create_xml()
    if grid.property_collection is None:
        grid.property_collection = rqp.PropertyCollection(support = grid)
    pc = grid.property_collection

    # shape of points property arrays
    points_shape = tuple(list(extent_kji) + [3])

    # create a static points property with multiple realisations
    for r in range(ensemble_size):
        stress = vec.unit_vectors(np.random.random(points_shape) + 0.1)
        pc.add_cached_array_to_imported_list(stress,
                                             'random stress vectors',
                                             'stress direction',
                                             uom = 'm',
                                             property_kind = 'length',
                                             realization = r,
                                             indexable_element = 'cells',
                                             points = True)
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()

    #  create a dynamic points property (indexable cells) related to the geological time series
    for r in range(ensemble_size):
        centres = grid.centre_point().copy()
        for time_index in range(time_series_size):
            pc.add_cached_array_to_imported_list(centres,
                                                 'dynamic cell centres',
                                                 'centres',
                                                 uom = 'm',
                                                 property_kind = 'length',
                                                 realization = r,
                                                 time_index = time_index,
                                                 indexable_element = 'cells',
                                                 points = True)
            centres[..., 2] += 100.0 * (r + 1)
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model(time_series_uuid = ts_uuid)

    #  create a dynamic points property (indexable nodes) related to the geological time series
    # also create a parallel set of active cell properties
    for r in range(ensemble_size):
        nodes = grid.points_ref().copy()
        for time_index in range(time_series_size):
            pc.add_cached_array_to_imported_list(nodes,
                                                 'dynamic nodes',
                                                 'points',
                                                 uom = 'm',
                                                 property_kind = 'length',
                                                 realization = r,
                                                 time_index = time_index,
                                                 indexable_element = 'nodes',
                                                 points = True)
            nodes[..., 2] += 100.0 * (r + 1)
            active_array = np.ones(extent_kji, dtype = bool)
            # de-activate some cells
            if extent_kji[0] > 1 and time_index < time_series_size // 2:
                active_array[extent_kji[0] // 2:] = False
            if extent_kji[2] > 1:
                active_array[:, :, r % extent_kji[2]] = False
            rqp.write_hdf5_and_create_xml_for_active_property(model,
                                                              active_array,
                                                              grid.uuid,
                                                              title = 'ACTIVE',
                                                              realization = r,
                                                              time_series_uuid = ts_uuid,
                                                              time_index = time_index)
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model(time_series_uuid = ts_uuid)

    # store the xml
    model.store_epc()

    # create a second grid being a copy of the first grid but with a fault
    fault_pillar_dict = {'the fault': [(2, 0), (2, 1), (2, 2), (1, 2), (1, 3), (1, 4)]}
    rqdm.add_faults(epc, source_grid = grid, full_pillar_list_dict = fault_pillar_dict, new_grid_title = 'faulted grid')

    # re-open the model and build a dynamic points property for the faulted grid
    # NB. this test implicitly has the oldest geometry as the 'official' grid geometry whereas
    # the RESQML documentation states that the geometry stored for the grid should be 'current' (ie. youngest)
    model = rq.Model(epc)
    f_grid = model.grid(title = 'faulted grid')
    assert f_grid is not None
    pc = f_grid.property_collection
    for r in range(faulted_ensemble_size):
        nodes = f_grid.points_ref().copy()
        for time_index in range(time_series_size):
            pc.add_cached_array_to_imported_list(nodes,
                                                 'faulted dynamic nodes',
                                                 'faulted points',
                                                 uom = 'm',
                                                 property_kind = 'length',
                                                 realization = r,
                                                 time_index = time_index,
                                                 indexable_element = 'nodes',
                                                 points = True)
            nodes[..., 2] += 101.0 * (r + 1)  # add more and more depth for each realisation
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model(time_series_uuid = ts_uuid)

    # store the xml again
    model.store_epc()

    # re-open the model and access the grid properties
    model = rq.Model(epc)
    grid = model.grid(title = 'unfaulted grid')
    pc = grid.property_collection

    # check that the grid as stored has all cells active
    assert grid.inactive is None or np.count_nonzero(grid.inactive) == 0

    # load a static points property for a single realisation
    sample_stress_part = pc.singleton(realization = ensemble_size // 2,
                                      citation_title = 'stress direction',
                                      points = True)
    assert sample_stress_part is not None
    sample_stress = pc.single_array_ref(realization = ensemble_size // 2,
                                        citation_title = 'stress direction',
                                        points = True)
    assert sample_stress is not None and sample_stress.ndim == 4
    assert np.all(sample_stress.shape[:3] == extent_kji)
    assert sample_stress.shape[3] == 3
    assert np.count_nonzero(np.isnan(sample_stress)) == 0
    stress_uuid = pc.uuid_for_part(sample_stress_part)

    # load the same property using the Property class
    stress_p = rqp.Property(model, uuid = stress_uuid)
    assert stress_p is not None
    assert stress_p.is_points()
    sample_stress = stress_p.array_ref()
    assert sample_stress is not None and sample_stress.ndim == 4

    # select the dynamic points properties related to the geological time series and indexable by cells
    cc = rqp.selective_version_of_collection(grid.property_collection,
                                             indexable = 'cells',
                                             points = True,
                                             time_series_uuid = ts_uuid)
    assert cc.number_of_parts() == ensemble_size * time_series_size

    #  check that 5 dimensional numpy arrays can be set up, each covering time indices for a single realisation
    for r in range(ensemble_size):
        rcc = rqp.selective_version_of_collection(cc, realization = r)
        assert rcc.number_of_parts() == time_series_size
        dynamic_nodes = rcc.time_series_array_ref()
        assert dynamic_nodes.ndim == 5
        assert dynamic_nodes.shape[0] == time_series_size
        assert np.all(dynamic_nodes.shape[1:4] == np.array(extent_kji))
        assert dynamic_nodes.shape[-1] == 3
        mean_depth = np.nanmean(dynamic_nodes[..., 2])
        assert mean_depth > 1000.0

    # select the dynamic points properties related to the geological time series and indexable by nodes
    nc = rqp.selective_version_of_collection(grid.property_collection,
                                             indexable = 'nodes',
                                             points = True,
                                             time_series_uuid = ts_uuid)
    assert nc.number_of_parts() == ensemble_size * time_series_size

    # check that the cached points for the grid can be populated from a points property
    grid.make_regular_points_cached()
    r = ensemble_size // 2
    ti = time_series_size // 2
    a = nc.single_array_ref(realization = r, time_index = ti)
    grid.set_cached_points_from_property(property_collection = nc,
                                         realization = r,
                                         time_index = ti,
                                         set_inactive = True,
                                         active_collection = grid.property_collection)
    assert_array_almost_equal(grid.points_cached, nc.single_array_ref(realization = r, time_index = ti))
    # check that grid's time index has been set
    assert grid.time_index == ti
    assert bu.matching_uuids(grid.time_series_uuid, ts_uuid)
    # and that the inactive array now indicates some cells are inactive
    assert grid.inactive is not None and np.count_nonzero(grid.inactive) > 0

    #  check that 5 dimensional numpy arrays can be set up, each covering realisations for a single time index
    for ti in range(time_series_size):
        tnc = rqp.selective_version_of_collection(nc, time_index = ti)
        assert tnc.number_of_parts() == ensemble_size
        dynamic_nodes = tnc.realizations_array_ref()
        assert dynamic_nodes.ndim == 5
        assert dynamic_nodes.shape[0] == ensemble_size
        assert np.all(dynamic_nodes.shape[1:4] == np.array(extent_kji) + 1)
        assert dynamic_nodes.shape[-1] == 3
        mean_depth = np.nanmean(dynamic_nodes[..., 2])
        assert mean_depth > 1000.0

    # check the faulted grid properties
    f_grid = model.grid(title = 'faulted grid')
    assert f_grid is not None
    assert not f_grid.k_direction_is_down
    f_grid.set_k_direction_from_points()
    assert not f_grid.k_direction_is_down

    # compute and cache cell centre points
    older_centres = f_grid.centre_point()
    assert hasattr(f_grid, 'array_centre_point') and f_grid.array_centre_point is not None

    # select the dynamic points properties related to the geological time series and indexable by nodes
    fnc = rqp.selective_version_of_collection(f_grid.property_collection,
                                              indexable = 'nodes',
                                              points = True,
                                              time_series_uuid = ts_uuid)
    assert fnc.number_of_parts() == faulted_ensemble_size * time_series_size

    # check that the cached points for the faulted grid can be populated from a points property
    r = faulted_ensemble_size // 2
    ti = time_series_size - 1
    p_uuid = fnc.uuid_for_part(fnc.singleton(realization = r, time_index = ti))
    assert p_uuid is not None
    f_grid.set_cached_points_from_property(points_property_uuid = p_uuid,
                                           set_grid_time_index = True,
                                           set_inactive = False)
    assert_array_almost_equal(f_grid.points_cached, fnc.single_array_ref(realization = r, time_index = ti))

    # check that grid's time index has been set
    assert f_grid.time_index == ti
    assert bu.matching_uuids(f_grid.time_series_uuid, ts_uuid)
    # and that the centre point cache has been invalidated
    assert not hasattr(f_grid, 'array_centre_point') or f_grid.array_centre_point is None

    # re-compute centre points based on the dynamically loaded geometry
    younger_centres = f_grid.centre_point()
    assert hasattr(f_grid, 'array_centre_point') and f_grid.array_centre_point is not None

    # check that later centres are vertically below the earlier centres
    # (in this example, the depths of all cells are increasing with time)
    assert_array_almost_equal(older_centres[..., :2], younger_centres[..., :2])  # xy
    assert np.all(older_centres[..., 2] < younger_centres[..., 2])  # depths


def test_remove_part_from_dict(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    assert pc is not None
    assert len(pc.parts()) == 8
    part = pc.parts()[0]

    # Act
    pc.remove_part_from_dict(part)

    # Assert
    assert len(pc.parts()) == 7
    assert part not in pc.parts()


def test_part_str(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection
    assert pc is not None
    part_disc = pc.parts()[0]
    part_cont = pc.parts()[-1]
    part_facet = pc.parts()[4]

    # Act / Assert
    assert pc.part_str(part_disc) == 'discrete (Zone)'
    assert pc.part_str(part_disc, include_citation_title = False) == 'discrete'
    assert pc.part_str(part_cont) == 'saturation: water; timestep: 2 (SW)'
    assert pc.part_str(part_cont, include_citation_title = False) == 'saturation: water; timestep: 2'
    assert pc.part_str(part_facet) == 'permeability rock: J (Perm)'
    assert pc.part_str(part_facet, include_citation_title = False) == 'permeability rock: J'


def test_part_filename(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection
    assert pc is not None
    part_disc = pc.parts()[0]
    part_cont = pc.parts()[-1]
    part_facet = pc.parts()[4]

    # Act / Assert
    assert pc.part_filename(part_disc) == 'discrete'
    assert pc.part_filename(part_cont) == 'saturation_water_ts_2'
    assert pc.part_filename(part_facet) == 'permeability_rock_J'


def test_grid_for_part(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    assert pc is not None
    part = pc.parts()[0]

    # Act
    grid = pc.grid_for_part(part)

    # Assert
    assert grid == model.grid()


def test_all_discrete(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    assert pc is not None

    # Act / Assert
    assert not pc.all_discrete()

    # Arrange
    for part in pc.parts():
        if pc.continuous_for_part(part):
            pc.remove_part_from_dict(part)

    # Act / Assert
    assert len(pc.parts()) == 4
    assert pc.all_discrete()


def test_h5_slice(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    assert pc is not None

    # Act  / Assert
    part = pc.parts()[0]
    full = pc.cached_part_array_ref(part)

    slice = pc.h5_slice(part, (0, 0))
    assert_array_almost_equal(slice, full[0, 0])

    slice = pc.h5_slice(part, (-1, -1))
    assert_array_almost_equal(slice, full[-1, -1])


def test_h5_overwrite_slice(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    assert pc is not None
    part = pc.parts()[0]

    # Act
    slice = pc.h5_slice(part, (0, 0))
    new_slice = np.zeros(shape = slice.shape)
    pc.h5_overwrite_slice(part, array_slice = new_slice, slice_tuple = (0, 0))

    # Assert
    new_full = pc.cached_part_array_ref(part)
    assert_array_almost_equal(new_slice, new_full[0, 0])


def test_string_lookup_for_part(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection
    lookup = model.parts_list_of_type('obj_StringTableLookup')[0]
    assert lookup is not None
    facies_part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Facies']
    assert len(facies_part) == 1

    # Act
    lookup_uuid = pc.string_lookup_uuid_for_part(facies_part[0])

    # Assert
    assert bu.matching_uuids(lookup_uuid, model.uuid_for_part(lookup))


def test_establish_has_multiple_realisations(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection
    # Assert initial model has multiple
    assert pc.establish_has_multiple_realizations()
    assert pc.has_multiple_realizations()
    # Remove parts with realiations
    for part in pc.parts():
        if pc.realization_for_part(part) is not None:
            pc.remove_part_from_dict(part)
    # Assert new model has not got multiple
    assert len(pc.parts()) == 8
    assert not pc.establish_has_multiple_realizations()
    assert not pc.has_multiple_realizations()


def test_establish_has_multiple_realisations_single(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection
    # Remove parts with realiation is None or 0
    for part in pc.parts():
        if pc.realization_for_part(part) in [None, 0]:
            pc.remove_part_from_dict(part)
    # Assert new model has not got multiple
    assert len(pc.parts()) == 2
    assert not pc.establish_has_multiple_realizations()
    assert not pc.has_multiple_realizations()


def test_establish_time_set_kind(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection
    # Assert model is not a time set
    assert pc.establish_time_set_kind() == 'not a time set'

    # Remove parts where ts != 2000-01-01Z
    for part in pc.parts():
        if pc.time_index_for_part(part) != 0:
            pc.remove_part_from_dict(part)
    # Assert new model is a single time
    assert len(pc.parts()) == 1
    assert pc.establish_time_set_kind() == 'single time'


def test_discombobulated_combobulated_face_arrays(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    orig = np.array([[[[1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6]]]])
    pc = model.grid().property_collection

    # Act
    combob = pc.combobulated_face_array(orig)
    assert combob.shape[-1] == 2
    assert combob.shape[-2] == 3
    discombob = pc.discombobulated_face_array(combob)
    assert discombob.shape[-1] == 6

    # Assert
    assert_array_almost_equal(orig, discombob)


def test_time_series_array_ref(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection
    # Trim the model to only contain properties with timesteps
    for part in pc.parts():
        if pc.time_index_for_part(part) is None:
            pc.remove_part_from_dict(part)
    # Make sure the number of properties is as expected, and they are all the same property
    assert len(pc.parts()) == 3
    assert len(set([pc.citation_title_for_part(part) for part in pc.parts()])) == 1
    # Pull out the full arrays to generate the expected output
    sw1, sw2, sw3 = [pc.cached_part_array_ref(part) for part in pc.parts()]
    expected = np.array([sw1, sw2, sw3])

    # Act
    output = pc.time_series_array_ref()

    # Assert
    assert_array_almost_equal(expected, output)


def test_inherit_parts_from_other_collection(example_model_with_prop_ts_rels):
    # Arrange
    copy_from = example_model_with_prop_ts_rels
    pc_from = copy_from.grid().property_collection

    pc_to = rqp.PropertyCollection()
    pc_to.set_support(model = copy_from, support_uuid = copy_from.grid().uuid)

    orig_from = len(pc_from.parts())

    # Act
    pc_to.inherit_parts_from_other_collection(pc_from)
    # Assert
    assert len(pc_to.parts()) == orig_from


def test_similar_parts_for_time_series_from_other_collection(example_model_with_prop_ts_rels):
    # Arrange
    copy_from = example_model_with_prop_ts_rels
    pc_from = copy_from.grid().property_collection

    pc_to = rqp.PropertyCollection()
    pc_to.set_support(model = copy_from, support_uuid = copy_from.grid().uuid)

    sw_parts = [part for part in pc_from.parts() if pc_from.citation_title_for_part(part) == 'SW']
    example_part = sw_parts[0]

    # Act
    pc_to.inherit_similar_parts_for_time_series_from_other_collection(other = pc_from, example_part = example_part)
    # Assert
    assert len(pc_to.parts()) == len(sw_parts)


def test_similar_parts_for_realizations_from_other_collection(example_model_with_prop_ts_rels):
    # Arrange
    copy_from = example_model_with_prop_ts_rels
    pc_from = copy_from.grid().property_collection

    pc_to = rqp.PropertyCollection(realization = 1)
    pc_to.set_support(model = copy_from)

    rel1_parts = [part for part in pc_from.parts() if pc_from.realization_for_part(part) == 1]
    example_part = rel1_parts[0]

    # Act
    pc_to.inherit_similar_parts_for_realizations_from_other_collection(other = pc_from, example_part = example_part)
    # Assert
    assert len(pc_to.parts()) == len(rel1_parts)


def test_property_over_time_series_from_collection(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection

    sw_parts = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'SW']
    example_part = sw_parts[0]

    # Act
    new_pc = rqp.property_over_time_series_from_collection(collection = pc, example_part = example_part)
    # Assert
    assert len(new_pc.parts()) == len(sw_parts)


def test_property_for_keyword_from_collection(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection

    sw_parts = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'SW']

    # Act
    new_pc = rqp.property_collection_for_keyword(collection = pc, keyword = 'sw')
    # Assert
    assert len(new_pc.parts()) == len(sw_parts)


def test_stringlookup_add_str(example_model_and_crs):
    # Arrange
    model, _ = example_model_and_crs
    lookup = rqp.StringLookup(parent_model = model)
    assert lookup.str_dict == {}
    # Act
    lookup.set_string(0, 'channel')
    # Assert
    assert lookup.str_dict == {0: 'channel'}


def test_create_property_set_xml(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    num_parts = len(pc.parts())

    # Act
    pc.create_property_set_xml('Grid property collection')
    model.store_epc()
    reload = rq.Model(model.epc_file)
    # Assert
    assert len(reload.parts_list_of_type('obj_PropertySet')) == 1

    # Act
    prop_set_root = reload.root_for_part(reload.parts_list_of_type('obj_PropertySet')[0])
    pset = rqp.PropertyCollection()
    pset.set_support(support = model.grid())
    pset.populate_from_property_set(prop_set_root)
    # Assert
    assert len(pset.parts()) == num_parts


def test_override_min_max(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    part = pc.parts()[0]

    print(pc.dict[part])
    vmin = pc.minimum_value_for_part(part)
    vmax = pc.maximum_value_for_part(part)

    # Act
    pc.override_min_max(part, min_value = vmin - 1, max_value = vmax + 1)
    # Assert
    assert pc.minimum_value_for_part(part) == vmin - 1
    assert pc.maximum_value_for_part(part) == vmax + 1


def test_set_support_mesh(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs

    ni = 5
    nj = 5
    origin = (0, 0, 0)
    di = dj = 50.0

    # make a regular mesh representation
    support = rqs.Mesh(model,
                       crs_uuid = crs.uuid,
                       mesh_flavour = 'regular',
                       ni = ni,
                       nj = nj,
                       origin = origin,
                       dxyz_dij = np.array([[di, 0.0, 0.0], [0.0, dj, 0.0]]),
                       title = 'regular mesh',
                       originator = 'Emma',
                       extra_metadata = {'testing mode': 'automated'})
    assert support is not None
    support.write_hdf5()
    support.create_xml()

    pc = rqp.PropertyCollection()
    pc.set_support(support = support)
    assert pc.support_uuid == support.uuid


# Set up expected arrays for normalized array tests
array1 = np.array([[[0, 1, 0, 1, 0], [1, 0, 1, 0, 1], [0, 1, 0, 1, 0], [1, 0, 1, 0, 1], [0, 1, 0, 1, 0]]] * 3)

array2 = np.where(array1 == 0, 0.5, array1)

array3 = np.array([[[1, 10, 10, 100, 100], [1, 10, 10, 100, 100], [1, 10, 10, 100, 100], [1, 10, 10, 100, 100],
                    [1, 10, 10, 100, 100]]] * 3)

array4 = np.where(array3 == 1, 0, array3)
array4 = np.where(array4 == 100, 1, array4)
array4 = np.where(array4 == 10, 0.090909, array4)

array5 = np.where(array3 == 1, 0, array3)
array5 = np.where(array5 == 100, 1, array5)
array5 = np.where(array5 == 10, 0.5, array5)

array6 = np.where(array3 == 100, np.nan, array3)
array6 = np.where(array6 == 1, 0, array6)
array6 = np.where(array6 == 10, 1, array6)

array7 = np.where(array3 == 10, np.nan, array3)
array7 = np.where(array7 == 1, 0, array7)
array7 = np.where(array7 == 100, 1, array7)


@pytest.mark.parametrize(
    'name,masked,log,discrete,trust,fix,array,emin,emax',
    [
        ('NTG', False, False, None, False, None, array1, 0, 0.5),  # Simple don't trust minmax
        ('NTG', False, False, None, True, None, array1, 0, 0.5),  # Simple trust minmax
        ('NTG', False, False, None, False, 0.5, array2, -0.5, 0.5),  # Fix 0 at 0.5
        ('Perm', False, False, None, False, None, array4, 1, 100),
        ('Perm', False, True, None, False, None, array5, 0, 2),
    ])  # Logarithmic
def test_norm_array_ref(example_model_with_properties, name, masked, log, discrete, trust, fix, array, emin, emax):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    cont = [part for part in pc.parts() if pc.citation_title_for_part(part) == name][0]

    # Act
    normed, vmin, vmax = pc.normalized_part_array(cont,
                                                  masked = masked,
                                                  use_logarithm = log,
                                                  discrete_cycle = discrete,
                                                  trust_min_max = trust,
                                                  fix_zero_at = fix)

    # Assert
    assert vmin == emin
    assert vmax == emax
    assert_array_almost_equal(normed, array)


def test_norm_array_ref_mask(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    grid = model.grid()
    # Set up a mask in the grid
    minimask = np.array([0, 0, 0, 1, 1])
    layermask = np.array([minimask] * 5)
    mask = np.array([layermask, layermask, layermask], dtype = 'bool')
    grid.inactive = mask
    pc = model.grid().property_collection
    cont = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    # Act
    normed, vmin, vmax = pc.normalized_part_array(cont, masked = True, use_logarithm = False)
    # Assert
    assert vmin == 1
    assert vmax == 10
    assert np.all(np.isclose(np.ma.array(array6, mask = np.isnan(array6)), normed))


def test_norm_array_ref_mask_equal(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    grid = model.grid()
    # Set up a mask in the grid
    minimask = np.array([1, 0, 0, 1, 1])
    layermask = np.array([minimask] * 5)
    mask = np.array([layermask, layermask, layermask], dtype = 'bool')
    grid.inactive = mask
    pc = model.grid().property_collection
    cont = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    # Act
    normed, vmin, vmax = pc.normalized_part_array(cont, masked = True, use_logarithm = False)
    # Assert
    assert vmin == 10
    assert vmax == 10
    assert_array_almost_equal(np.ones(shape = (3, 5, 5)) / 2, normed)


def test_norm_array_ref_log_mask(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    grid = model.grid()
    # Set up a mask in the grid
    minimask = np.array([0, 1, 1, 0, 0])
    layermask = np.array([minimask] * 5)
    mask = np.array([layermask, layermask, layermask], dtype = 'bool')
    grid.inactive = mask
    pc = model.grid().property_collection
    cont = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    # Act
    normed, vmin, vmax = pc.normalized_part_array(cont, masked = True, use_logarithm = True)
    # Assert
    assert vmin == 0
    assert vmax == 2
    assert np.all(np.isclose(np.ma.array(array7, mask = np.isnan(array7)), normed))


def test_normalized_part_array_discrete(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    disc = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Zone'][0]

    # Act
    normed, vmin, vmax = pc.normalized_part_array(disc, discrete_cycle = 3)
    # Assert
    assert vmin == 0
    assert vmax == 2
    assert normed[0, 0, 0] == 0.5
    assert normed[1, 0, 0] == 1
    assert normed[2, 0, 0] == 0


def test_create_xml_minmax_none(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    array = np.ones(shape = (3, 5, 5))
    array[0, 0, 0] = 2
    support_uuid = model.grid().uuid
    ext_uuid = model.h5_uuid()

    p_node = pc.create_xml(ext_uuid = ext_uuid,
                           property_array = array,
                           title = 'Tester',
                           property_kind = 'continuous',
                           support_uuid = support_uuid,
                           p_uuid = bu.new_uuid(),
                           uom = 'Euc',
                           add_min_max = True,
                           min_value = None,
                           max_value = None,
                           indexable_element = 'cells',
                           count = 1)

    assert rqet.find_tag_text(p_node, 'MinimumValue') == '1.0'
    assert rqet.find_tag_text(p_node, 'MaximumValue') == '2.0'


def test_create_xml_minmax_none_discrete(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    array = np.ones(shape = (3, 5, 5))
    array[0, 0, 0] = 2
    support_uuid = model.grid().uuid
    ext_uuid = model.h5_uuid()

    p_node = pc.create_xml(ext_uuid = ext_uuid,
                           property_array = array,
                           title = 'Tester',
                           property_kind = 'discrete',
                           support_uuid = support_uuid,
                           p_uuid = bu.new_uuid(),
                           uom = 'Euc',
                           discrete = True,
                           add_min_max = True,
                           min_value = None,
                           max_value = None,
                           indexable_element = 'cells',
                           count = 1)

    assert rqet.find_tag_text(p_node, 'MinimumValue') == '1'
    assert rqet.find_tag_text(p_node, 'MaximumValue') == '2'


def test_basic_static_property_parts_ntgnone(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act / Assert - Check it finds ntg initially
    ntg, por, permi, permj, permk = pc.basic_static_property_parts()
    assert ntg is not None
    assert pc.citation_title_for_part(ntg) == 'NTG'

    # Arrange - delete ntg
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'NTG'][0]
    pc.remove_part_from_dict(part)

    # Act / Assert - check it now finds nothing
    ntg, por, permi, permj, permk = pc.basic_static_property_parts()
    assert ntg is None


def test_basic_static_property_parts_pornone(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act / Assert - Check it finds por initially
    ntg, por, permi, permj, permk = pc.basic_static_property_parts()
    assert por is not None
    assert pc.citation_title_for_part(por) == 'POR'

    # Arrange - Delete por
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'POR'][0]
    pc.remove_part_from_dict(part)

    # Act / Assert - Check it now finds nothing
    ntg, por, permi, permj, permk = pc.basic_static_property_parts()
    assert por is None


def test_basic_static_property_parts_permnone(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act / Assert - Check it finds permi initially
    ntg, por, permi, permj, permk = pc.basic_static_property_parts()
    assert permi is not None
    assert permj is None
    assert permk is None
    assert pc.citation_title_for_part(permi) == 'Perm'

    # Arrange - Delete permi
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    pc.remove_part_from_dict(part)
    # Act / Assert - Check it now finds nothing
    ntg, por, permi, permj, permk = pc.basic_static_property_parts()
    assert permi is None
    assert permj is None
    assert permk is None


def test_basic_static_property_parts_permshared(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act
    ntg, por, permi, permj, permk = pc.basic_static_property_parts(share_perm_parts = True)

    # Assert
    assert permi is not None
    assert permj is not None
    assert permk is not None
    assert permi == permj == permk
    assert pc.citation_title_for_part(permi) == 'Perm'
    assert pc.facet_for_part(permi) == 'I'


@pytest.mark.parametrize('facet,expected_none', [('J', [True, False, True]), ('K', [True, True, False]),
                                                 ('IJ', [False, False, True]), ('IJK', [False, False, False]),
                                                 ('Invalid', [False, True, True])])
def test_basic_static_property_parts_perm_facet(example_model_with_properties, facet, expected_none):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    array = pc.cached_part_array_ref(part)
    pc.remove_part_from_dict(part)

    pc.add_cached_array_to_imported_list(cached_array = array,
                                         source_info = '',
                                         keyword = 'Testfacet',
                                         discrete = False,
                                         property_kind = 'permeability rock',
                                         facet_type = 'direction',
                                         facet = facet)
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()

    # Act
    _, _, permi, permj, permk = pc.basic_static_property_parts()

    # Assert
    for actual, expected in zip([permi, permj, permk], expected_none):
        if expected:
            assert actual is None
        else:
            assert actual is not None
            assert pc.citation_title_for_part(actual) == 'Testfacet'


@pytest.mark.parametrize('facet_list,expected_none,expected_names',
                         [(['K', 'I'], [False, True, False], ['Testfacet_I', None, 'Testfacet_K']),
                          (['K', 'J'], [True, False, False], [None, 'Testfacet_J', 'Testfacet_K']),
                          (['IJ', 'K'], [False, False, False], ['Testfacet_IJ', 'Testfacet_IJ', 'Testfacet_K']),
                          (['IJK'], [False, False, False], ['Testfacet_IJK', 'Testfacet_IJK', 'Testfacet_IJK']),
                          (['Invalid', 'K'], [True, True, False], [None, None, 'Testfacet_K'])])
def test_basic_static_property_parts_perm_multiple_facet(example_model_with_properties, facet_list, expected_none,
                                                         expected_names):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    array = pc.cached_part_array_ref(part)
    pc.remove_part_from_dict(part)
    for facet in facet_list:
        pc.add_cached_array_to_imported_list(cached_array = array,
                                             source_info = '',
                                             keyword = f'Testfacet_{facet}',
                                             discrete = False,
                                             property_kind = 'permeability rock',
                                             facet_type = 'direction',
                                             facet = facet)
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()

    # Act
    _, _, permi, permj, permk = pc.basic_static_property_parts()

    # Assert
    for actual, expected, name in zip([permi, permj, permk], expected_none, expected_names):
        if expected:
            assert actual is None, f'Expected none for {name}'
        else:
            assert actual is not None
            assert pc.citation_title_for_part(actual) == name


@pytest.mark.parametrize('name_list,expected_none', [(['KI', 'KJ', 'KK'], [False, False, False]),
                                                     (['KX', 'KY', 'KZ'], [False, False, False]),
                                                     (['PERMI', 'PERMJ', 'PERMK'], [False, False, False]),
                                                     (['PERMX', 'PERMY', 'PERMZ'], [False, False, False])])
def test_basic_static_property_parts_perm_multiple_name(example_model_with_properties, name_list, expected_none):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    array = pc.cached_part_array_ref(part)
    pc.remove_part_from_dict(part)
    for name in name_list:
        pc.add_cached_array_to_imported_list(cached_array = array,
                                             source_info = '',
                                             keyword = name,
                                             discrete = False,
                                             property_kind = 'permeability rock')
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()

    # Act
    _, _, permi, permj, permk = pc.basic_static_property_parts()

    # Assert
    for actual, expected, name in zip([permi, permj, permk], expected_none, name_list):
        if expected:
            assert actual is None, f'Expected none for {name}'
        else:
            assert actual is not None
            assert pc.citation_title_for_part(actual) == name


@pytest.mark.parametrize('name_list,facet', [(['KI', 'KX'], 'I'), (['KJ', 'KY'], 'J'), (['KZ', 'KK'], 'K')])
def test_basic_static_property_parts_perm_repeat(example_model_with_properties, name_list, facet):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    array = pc.cached_part_array_ref(part)
    pc.remove_part_from_dict(part)
    for name in name_list:
        pc.add_cached_array_to_imported_list(cached_array = array,
                                             source_info = '',
                                             keyword = name,
                                             discrete = False,
                                             facet_type = 'direction',
                                             facet = facet,
                                             property_kind = 'permeability rock')
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()

    # Act
    _, _, permi, permj, permk = pc.basic_static_property_parts()

    # Assert
    assert permi is None
    assert permj is None
    assert permk is None


def test_basic_static_property_parts_perm_options(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    array = pc.cached_part_array_ref(part)
    ntgpart = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'NTG'][0]
    ntgarray = pc.cached_part_array_ref(ntgpart)
    pc.add_cached_array_to_imported_list(cached_array = array,
                                         source_info = '',
                                         keyword = 'PermK',
                                         discrete = False,
                                         facet_type = 'direction',
                                         facet = 'J',
                                         property_kind = 'permeability rock')
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()

    # Act
    _, _, permi, permj, permk = pc.basic_static_property_parts(perm_k_mode = 'none')
    assert permi is not None
    assert permj is not None
    assert permk is None

    _, _, permi, permj, permk = pc.basic_static_property_parts(perm_k_mode = None)
    assert permi is not None
    assert permj is not None
    assert permk is None

    _, _, permi, permj, permk = pc.basic_static_property_parts(perm_k_mode = 'ratio', perm_k_ratio = 0.5)
    assert permi is not None
    assert permj is not None
    assert permk is not None
    karray = pc.cached_part_array_ref(permk)
    assert_array_almost_equal(karray, array * 0.5)


def test_basic_static_property_parts_perm_options_ntg(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    array = pc.cached_part_array_ref(part)
    ntgpart = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'NTG'][0]
    ntgarray = pc.cached_part_array_ref(ntgpart)
    pc.add_cached_array_to_imported_list(cached_array = array,
                                         source_info = '',
                                         keyword = 'PermK',
                                         discrete = False,
                                         facet_type = 'direction',
                                         facet = 'J',
                                         property_kind = 'permeability rock')
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()

    ntg, _, permi, permj, permk = pc.basic_static_property_parts(perm_k_mode = 'ntg', perm_k_ratio = 0.5)
    assert permi is not None
    assert permj is not None
    assert permk is not None
    karray = pc.cached_part_array_ref(permk)
    assert_array_almost_equal(karray, (array / 2) * ntgarray)


def test_basic_static_property_parts_perm_options_ntgsquared(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'Perm'][0]
    array = pc.cached_part_array_ref(part)
    ntgpart = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'NTG'][0]
    ntgarray = pc.cached_part_array_ref(ntgpart)
    pc.add_cached_array_to_imported_list(cached_array = array,
                                         source_info = '',
                                         keyword = 'PermK',
                                         discrete = False,
                                         facet_type = 'direction',
                                         facet = 'J',
                                         property_kind = 'permeability rock')
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()

    _, _, permi, permj, permk = pc.basic_static_property_parts(perm_k_mode = 'ntg squared', perm_k_ratio = 0.5)
    assert permi is not None
    assert permj is not None
    assert permk is not None
    karray = pc.cached_part_array_ref(permk)
    assert_array_almost_equal(karray, (array / 2) * (ntgarray * ntgarray))


@pytest.mark.parametrize('keyword,kind,facet_type,facet', [('bv', 'rock volume', 'netgross', 'gross'),
                                                           ('brv', 'rock volume', 'netgross', 'gross'),
                                                           ('pv', 'pore volume', None, None),
                                                           ('pvr', 'pore volume', None, None),
                                                           ('porv', 'pore volume', None, None),
                                                           ('mdep', 'depth', 'what', 'cell centre'),
                                                           ('depth', 'depth', 'what', 'cell top'),
                                                           ('tops', 'depth', 'what', 'cell top'),
                                                           ('mids', 'depth', 'what', 'cell centre'),
                                                           ('ntg', 'net to gross ratio', None, None),
                                                           ('netgrs', 'net to gross ratio', None, None),
                                                           ('netv', 'rock volume', 'netgross', 'net'),
                                                           ('nrv', 'rock volume', 'netgross', 'net'),
                                                           ('dzc', 'thickness', 'netgross', 'gross'),
                                                           ('dzn', 'thickness', 'netgross', 'net'),
                                                           ('dz', 'thickness', 'netgross', 'gross'),
                                                           ('dznet', 'thickness', 'netgross', 'net'),
                                                           ('dxc', 'cell length', 'direction', 'I'),
                                                           ('dyc', 'cell length', 'direction', 'J'),
                                                           ('dx', 'cell length', 'direction', 'I'),
                                                           ('dy', 'cell length', 'direction', 'J'),
                                                           ('dxaaa', 'length', 'direction', 'X'),
                                                           ('dyaaa', 'length', 'direction', 'Y'),
                                                           ('dzaaa', 'length', 'direction', 'Z'),
                                                           ('por', 'porosity', None, None),
                                                           ('poro', 'porosity', None, None),
                                                           ('porosity', 'porosity', None, None),
                                                           ('kh', 'permeability thickness', None, None),
                                                           ('transx', 'transmissibility', 'direction', 'I'),
                                                           ('tx', 'transmissibility', 'direction', 'I'),
                                                           ('ty', 'transmissibility', 'direction', 'J'),
                                                           ('tz', 'transmissibility', 'direction', 'K'),
                                                           ('ti', 'transmissibility', 'direction', 'I'),
                                                           ('tj', 'transmissibility', 'direction', 'J'),
                                                           ('tk', 'transmissibility', 'direction', 'K'),
                                                           ('p', 'pressure', None, None),
                                                           ('pressure', 'pressure', None, None),
                                                           ('sw', 'saturation', 'what', 'water'),
                                                           ('so', 'saturation', 'what', 'oil'),
                                                           ('sg', 'saturation', 'what', 'gas'),
                                                           ('satw', 'saturation', 'what', 'water'),
                                                           ('sato', 'saturation', 'what', 'oil'),
                                                           ('satg', 'saturation', 'what', 'gas'),
                                                           ('soil', 'saturation', 'what', 'oil'),
                                                           ('swl', 'saturation', 'what', 'water minimum'),
                                                           ('swr', 'saturation', 'what', 'water residual'),
                                                           ('sgl', 'saturation', 'what', 'gas minimum'),
                                                           ('sgr', 'saturation', 'what', 'gas residual'),
                                                           ('swro', 'saturation', 'what', 'water residual to oil'),
                                                           ('sgro', 'saturation', 'what', 'gas residual to oil'),
                                                           ('sor', 'saturation', 'what', 'oil residual'),
                                                           ('swu', 'saturation', 'what', 'water maximum'),
                                                           ('sgu', 'saturation', 'what', 'gas maximum'),
                                                           ('wip', 'fluid volume', 'what', 'water'),
                                                           ('oip', 'fluid volume', 'what', 'oil'),
                                                           ('gip', 'fluid volume', 'what', 'gas'),
                                                           ('mobw', 'fluid volume', 'what', 'water (mobile)'),
                                                           ('mobo', 'fluid volume', 'what', 'oil (mobile)'),
                                                           ('mobg', 'fluid volume', 'what', 'gas (mobile)'),
                                                           ('ocip', 'fluid volume', 'what', 'oil condensate'),
                                                           ('tmx', 'transmissibility multiplier', 'direction', 'I'),
                                                           ('tmy', 'transmissibility multiplier', 'direction', 'J'),
                                                           ('tmz', 'transmissibility multiplier', 'direction', 'K'),
                                                           ('tmflx', 'transmissibility multiplier', 'direction', 'I'),
                                                           ('tmfly', 'transmissibility multiplier', 'direction', 'J'),
                                                           ('tmflz', 'transmissibility multiplier', 'direction', 'K'),
                                                           ('multx', 'transmissibility multiplier', 'direction', 'I'),
                                                           ('multy', 'transmissibility multiplier', 'direction', 'J'),
                                                           ('multz', 'transmissibility multiplier', 'direction', 'K'),
                                                           ('multbv', 'property multiplier', 'what', 'rock volume'),
                                                           ('multpv', 'property multiplier', 'what', 'pore volume'),
                                                           ('rs', 'solution gas-oil ratio', None, None),
                                                           ('rv', 'vapor oil-gas ratio', None, None),
                                                           ('temp', 'thermodynamic temperature', None, None),
                                                           ('temperature', 'thermodynamic temperature', None, None),
                                                           ('dad', 'code', 'what', 'dad'),
                                                           ('kid', 'code', 'what', 'inactive'),
                                                           ('unpack', 'code', 'what', 'unpack'),
                                                           ('deadcell', 'code', 'what', 'inactive'),
                                                           ('inactive', 'code', 'what', 'inactive'),
                                                           ('livecell', 'active', None, None),
                                                           ('act test', 'active', None, None),
                                                           ('ireg', 'region initialization', None, None),
                                                           ('region', 'region initialization', None, None),
                                                           ('cregion', 'region initialization', None, None),
                                                           ('uid', 'index', 'what', 'uid'),
                                                           ('noneoftheabove', None, None, None)])
def test_property_kind_and_facet_from_keyword(keyword, kind, facet_type, facet):
    out_kind, out_type, out_facet = property_kind_and_facet_from_keyword(keyword)
    assert out_kind == kind
    assert out_type == facet_type
    assert out_facet == facet


@pytest.mark.parametrize('expected,xy_uom,z_uom,property_kind,minimum,maximum,facet_type,facet',
                         [('m3', 'm', 'm', 'rock volume', None, None, None, None),
                          ('m3', 'm', 'm', 'volume', None, None, None, None),
                          ('ft3', 'ft', 'ft', 'rock volume', None, None, None, None),
                          ('ft3', 'ft', 'ft', 'volume', None, None, None, None),
                          ('m3', 'm', 'm', 'pore volume', None, None, None, None),
                          ('bbl', 'ft', 'ft', 'pore volume', None, None, None, None),
                          (None, 'm', 'ft', 'volume', None, None, None, None),
                          ('m', 'm', 'm', 'depth', None, None, None, None),
                          ('ft', 'ft', 'ft', 'depth', None, None, None, None),
                          ('m', 'ft', 'm', 'depth', None, None, None, None),
                          ('ft', 'm', 'ft', 'depth', None, None, None, None),
                          ('m', 'm', 'm', 'cell length', None, None, None, None),
                          ('ft', 'ft', 'ft', 'cell length', None, None, None, None),
                          ('%', 'm', 'ft', 'net to gross ratio', None, 50, None, None),
                          (None, 'm', 'm', 'net to gross ratio', None, -1, None, None),
                          ('m3/m3', 'm', 'm', 'net to gross ratio', None, 0.5, None, None),
                          ('ft3/ft3', 'ft', 'ft', 'net to gross ratio', None, 0.5, None, None),
                          ('Euc', 'm', 'ft', 'net to gross ratio', None, None, None, None),
                          ('%', 'm', 'm', 'porosity', None, 50, None, None),
                          ('%', 'm', 'm', 'saturation', None, 50, None, None),
                          ('mD', 'm', 'm', 'permeability rock', None, None, None, None),
                          ('mD', 'm', 'm', 'rock permeability', None, None, None, None),
                          ('mD.m', 'm', 'm', 'permeability thickness', None, None, None, None),
                          ('mD.ft', 'ft', 'ft', 'permeability thickness', None, None, None, None),
                          ('mD.m', 'm', 'm', 'permeability length', None, None, None, None),
                          ('mD.ft', 'ft', 'ft', 'permeability length', None, None, None, None),
                          (None, 'ft', 'm', 'permeability length', None, None, None, None),
                          ('m3', 'm', 'm', 'fluid volume', None, None, None, None),
                          ('1000 ft3', 'ft', 'ft', 'fluid volume', None, None, 'what', 'gas'),
                          ('bbl', 'ft', 'ft', 'fluid volume', None, None, 'what', 'oil'),
                          ('bbl', 'ft', 'ft', 'fluid volume', None, None, None, None),
                          ('m3.cP/(kPa.d)', 'm', 'm', 'transmissibility', None, None, None, None),
                          ('bbl.cP/(psi.d)', 'ft', 'ft', 'transmissibility', None, None, None, None),
                          ('kPa', 'm', 'm', 'pressure', None, None, None, None),
                          ('psi', 'ft', 'ft', 'pressure', None, None, None, None),
                          (None, 'ft', 'm', 'pressure', None, 0, None, None),
                          ('kPa', 'ft', 'm', 'pressure', None, 20000, None, None),
                          ('bar', 'ft', 'm', 'pressure', None, 450, None, None),
                          ('psi', 'ft', 'm', 'pressure', None, 4500, None, None),
                          (None, 'ft', 'm', 'pressure', None, None, None, None),
                          ('m3/m3', 'm', 'm', 'solution gas-oil ratio', None, None, None, None),
                          ('1000 ft3/bbl', 'ft', 'ft', 'solution gas-oil ratio', None, None, None, None),
                          ('m3/m3', 'm', 'm', 'vapor oil-gas ratio', None, None, None, None),
                          ('0.001 bbl/ft3', 'ft', 'ft', 'vapor oil-gas ratio', None, None, None, None),
                          ('Euc', 'm', 'm', 'some kind of multiplier', None, None, None, None),
                          (None, 'm', 'm', 'none of the above', None, None, None, None)])
def test_guess_uom(tmp_model, expected, xy_uom, z_uom, property_kind, minimum, maximum, facet_type, facet):
    model = tmp_model
    crs = Crs(parent_model = model, z_inc_down = True, xy_units = xy_uom, z_units = z_uom)
    crs.create_xml()
    support = grr.RegularGrid(parent_model = model,
                              origin = (0, 0, 0),
                              extent_kji = (3, 5, 5),
                              crs_uuid = rqet.uuid_for_part_root(model.crs_root),
                              set_points_cached = True)
    support.cache_all_geometry_arrays()
    support.write_hdf5_from_caches(file = model.h5_file_name(file_must_exist = False), mode = 'w')

    support.create_xml(ext_uuid = model.h5_uuid(),
                       title = 'grid',
                       write_geometry = True,
                       add_cell_length_properties = False)
    model.store_epc()

    uom = guess_uom(property_kind,
                    minimum = minimum,
                    maximum = maximum,
                    support = support,
                    facet_type = facet_type,
                    facet = facet)
    assert uom == expected


def test_property_kind_list(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act
    element = pc.property_kind_list()

    # Assert
    assert element == ['discrete', 'net to gross ratio', 'permeability rock', 'porosity', 'saturation']


def test_indexable_list(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act
    element = pc.unique_indexable_element_list()

    # Assert
    assert element == ['cells']


def test_facet_type_list(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act
    element = pc.facet_type_list()

    # Assert
    assert element == ['direction']


def test_facet_list(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act
    element = pc.facet_list()

    # Assert
    assert element == ['I']


def test_time_series_uuid_list(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection
    ts_uuid = [model.uuid_for_part(part) for part in model.parts_list_of_type('obj_TimeSeries')]

    # Act
    element = pc.time_series_uuid_list()

    # Assert
    assert element == ts_uuid


def test_uom_list(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act
    element = pc.uom_list()

    # Assert
    assert element == ['m3/m3', 'mD']


def test_string_lookup_uuid_list(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    pc = model.grid().property_collection
    lookup_uuid = [model.uuid_for_part(part) for part in model.parts_list_of_type('obj_StringTableLookup')]

    # Act
    element = pc.string_lookup_uuid_list()

    # Assert
    assert element == lookup_uuid


def test_shape_and_type_of_part(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = model.grid().property_collection

    # Act / Assert
    for part in pc.parts():
        shape, dtype = pc.shape_and_type_of_part(part)
        assert shape == (3, 5, 5)
        if pc.continuous_for_part(part):
            assert dtype == '<f8'
        else:
            assert dtype == 'int32'


def test_well_interval_property_collection(example_model_and_cellio):
    # Arrange
    model, cellio_file, well_name = example_model_and_cellio
    grid = model.grid()
    bw = rqw.BlockedWell(model, use_face_centres = True)
    bw.import_from_rms_cellio(cellio_file = cellio_file, well_name = well_name, grid = grid)
    bw.write_hdf5()
    bw.create_xml()
    model.store_epc()

    pc = rqp.PropertyCollection(support = bw)

    # Add parts to collection
    zones = np.array([1, 2, 3])
    sw = np.array([0.1, 0.2, 1])
    por = np.array([0, 0.3, 0.24])
    for array, keyword, discrete in zip([zones, sw, por], ['Zone', 'sw', 'por'], [True, False, False]):
        pc.add_cached_array_to_imported_list(array,
                                             source_info = 'testing data',
                                             keyword = keyword,
                                             discrete = discrete)
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()
    model.store_epc()

    # Reload the model
    reload = rq.Model(model.epc_file)
    loadbw = rqw.BlockedWell(parent_model = reload, uuid = bw.uuid)

    # Act
    wipc = rqp.WellIntervalPropertyCollection(frame = loadbw)
    df = wipc.to_pandas()

    # Assert
    assert len(list(wipc.logs())) == 3
    assert list(df.columns) == ['Zone', 'sw', 'por']
    assert all(df['Zone'].values == zones)
    assert all(df['por'].values == por)
    assert all(df['sw'].values == sw)

    for log in wipc.logs():
        if log.name == 'Zone':
            assert all(log.values() == zones)
        elif log.name == 'sw':
            assert all(log.values() == sw)
        else:
            assert all(log.values() == por)


def test_load_grid_property_collection(example_model_with_prop_ts_rels):
    # Arrange
    model = example_model_with_prop_ts_rels
    grid = model.grid()

    # Act
    pc = rqp.GridPropertyCollection(grid = grid)

    # Assert
    assert pc is not None
    assert len(pc.parts()) == 12


def test_write_read_nexus_array(example_model_with_prop_ts_rels, tmp_path):
    # Arrange
    model = example_model_with_prop_ts_rels
    grid = model.grid()
    directory = str(tmp_path)

    # Act
    pc = rqp.GridPropertyCollection(grid = grid)
    assert pc is not None
    numparts = len(pc.parts())
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'SW'][0]
    pc.write_nexus_property_generating_filename(part = part, directory = directory)

    outfile = os.path.join(tmp_path, 'SW_what_water_t_0')
    assert os.path.exists(outfile)

    with open(outfile, 'r') as f:
        lines = f.readlines()
        assert lines[1] == '! Extent of array is: [5, 5, 3]\n'
        assert lines[-1] == '1.000\t0.500\t1.000\t0.500\t1.000\n'

    _ = pc.import_nexus_property_to_cache(file_name = outfile, keyword = 'Sw1', discrete = False)
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()
    model.store_epc()

    reload = rq.Model(model.epc_file)
    newpc = rqp.GridPropertyCollection(grid = reload.grid())
    assert len(newpc.parts()) == numparts + 1


def test_slice_for_box(example_model_with_properties):
    # Arrange
    model = example_model_with_properties
    pc = rqp.GridPropertyCollection(grid = model.grid())
    assert pc is not None
    part = [part for part in pc.parts() if pc.citation_title_for_part(part) == 'SW'][0]

    # Act
    myslice = pc.h5_slice_for_box(part = part, box = np.array([[0, 0, 0], [1, 2, 2]]))

    expected = np.array([[[1, 0.5, 1], [1, 0.5, 1], [1, 0.5, 1]], [[1, 0.5, 1], [1, 0.5, 1], [1, 0.5, 1]]])

    # Assert
    assert_array_almost_equal(expected, myslice)



# test hexahedral volume calculations

import math as maths

import numpy as np
from numpy.random import random
from numpy.testing import assert_array_almost_equal

import resqpy.olio.volume as vol


def unit_cube():
    cp = np.zeros((2, 2, 2, 3))  # pagoda style corner point array for a single cell
    # set to unit cube with xyz aligned with ijk respectively
    cp[:, :, 1, 0] = 1.0
    cp[:, 1, :, 1] = 1.0
    cp[1, :, :, 2] = 1.0
    return cp


def test_one_cell_volume():
    cp = unit_cube().copy()
    assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 1.0), 'unit cube cell volume calculation fails'
    # stretch in x (I)
    cp[:, :, 1, 0] *= 2.5
    assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 2.5), 'square prism cell volume calculation fails'
    # stretch in y & z (J & K)
    cp[:, 1, :, 1] *= 2.0
    cp[1, :, :, 2] *= 3.0
    assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False),
                         15.0), 'regular cuboid cell volume calculation fails'
    # apply a shear in x & y (should not change volume)
    cp[1, :, :, :2] += 0.5
    assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False),
                         15.0), 'parallelepiped cell volume calculation fails'
    # translate into negative coordinate realm
    cp[:] -= 3.827
    assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 15.0), 'translated cell volume calculation fails '
    # invert one axis
    cp[:, :, :, 1] *= -1.0
    assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = True),
                         15.0), 'off hand True cell volume calculation fails'
    # collapse to plane
    cp[:, :, :, 2] = 0.0
    assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 0.0), 'planar cell volume calculation fails'
    assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = True),
                         0.0), 'off hand True planar cell volume calculation fails'
    # test wedge
    cp = unit_cube().copy()
    cp[1, 1, :, 2] = 0.0
    assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 0.5), 'triangular prism volume calculation fails'


def test_array_volume():
    # note: the multiple cells generated here are all overlapping, which would not usually be the case
    cp = np.zeros((5, 10, 15, 2, 2, 2, 3))  # pagoda style corner point array
    cp[:] = unit_cube().reshape((1, 1, 1, 2, 2, 2, 3)) * 2.0
    cp[:] += random(cp.size).reshape(cp.shape) - 0.5
    # move whole cells by random translations
    translate = (random(cp.size // 8) * 123.2356).reshape((cp.shape[0], cp.shape[1], cp.shape[2], 1, 1, 1, 3))
    cp[:] += translate
    # compute volumes using two different routines in the volume module and make sure they give the same answer
    v1 = vol.tetra_volumes_slow(cp, off_hand = False)
    v2 = vol.tetra_volumes(cp, off_hand = False)
    # following is a stringent test requiring that exactly the same mathematical operations have been performed
    assert_array_almost_equal(v1, v2)

    # the random changes to corner points should have left the volumes within a certain range
    assert np.all(v1 >= 0.0), 'negative volume(s) returned by array function'
    assert np.all(v1 <= 27.0), 'exaggerated volume(s) returned by array function'
    # translate again and check that volumes are unchanged
    translate = (random(cp.size // 8) * 3.23478).reshape((cp.shape[0], cp.shape[1], cp.shape[2], 1, 1, 1, 3))
    cp[:] -= translate
    v2 = vol.tetra_volumes(cp, off_hand = False)
    assert_array_almost_equal(v1, v2)

    # test handedness inversion
    cp[:, :, :, :, :, :, 0] *= -1.0
    v1 = vol.tetra_volumes_slow(cp, off_hand = True)
    v2 = vol.tetra_volumes(cp, off_hand = True)
    assert_array_almost_equal(v1, v2)
    assert np.all(v1 >= 0.0), 'negative volume(s) returned by array function'
    assert np.all(v1 <= 27.0), 'exaggerated volume(s) returned by array function'


def test_pyramid_volume():
    apex = (1.0, 1.0, -3.0)
    a = (0.0, 0.0, 0.0)
    b = (0.0, 2.0, 0.0)
    c = (2.0, 2.0, 0.0)
    d = (2.0, 0.0, 0.0)
    assert maths.isclose(vol.pyramid_volume(apex, a, b, c, d), 4.0)
    assert maths.isclose(vol.pyramid_volume(apex, a, d, c, b, crs_is_right_handed = True), 4.0)


def test_tetrahedron_volume():
    one_over_root_two = 1.0 / maths.sqrt(2.0)
    a = (-1.0, 0.0, -one_over_root_two)
    b = (1.0, 0.0, -one_over_root_two)
    c = (0.0, -1.0, one_over_root_two)
    d = (0.0, 1.0, one_over_root_two)
    assert maths.isclose(vol.tetrahedron_volume(a, b, c, d), 4.0 / (3.0 * maths.sqrt(2.0)))



"""point_inclusion.py: functions to test whether a point is within a polygon; also line intersection with planes."""

version = '29th April 2021'

import logging

log = logging.getLogger(__name__)

import math as maths
import os

import numpy as np

import resqpy.olio.simple_lines as sl


def pip_cn(p, poly):
    """2D point inclusion: returns True if point is inside polygon (uses crossing number algorithm).

    arguments:
       p (numpy array, tuple or list of at least 2 floats): xy point to test for inclusion in polygon
       poly (2D numpy array, tuple or list of tuples or lists of at least 2 floats): the xy points
          defining a polygon

    returns:
       boolean: True if point is within the polygon

    note:
       if the polygon is represented by a closed resqpy Polyline, pass Polyline.coordinates as poly
    """

    # p should be a tuple-like object with first two elements x, y
    # poly should be a numpy array with first axis being the different vertices
    #    and the second starting x, y

    crossings = 0

    for edge in range(len(poly)):
        v1 = poly[edge - 1]
        v2 = poly[edge]
        if ((v1[0] > p[0] or v2[0] > p[0]) and ((v1[1] <= p[1] and v2[1] > p[1]) or (v1[1] > p[1] and v2[1] <= p[1]))):
            if v1[0] > p[0] and v2[0] > p[0]:
                crossings += 1
            elif p[0] < v1[0] + (v2[0] - v1[0]) * (p[1] - v1[1]) / (v2[1] - v1[1]):
                crossings += 1

    return bool(crossings & 1)


def pip_wn(p, poly):
    """2D point inclusion: returns True if point is inside polygon (uses winding number algorithm).

    arguments:
       p (numpy array, tuple or list of at least 2 floats): xy point to test for inclusion in polygon
       poly (2D numpy array, tuple or list of tuples or lists of at least 2 floats): the xy points
          defining a polygon

    returns:
       boolean: True if point is within the polygon

    note:
       if the polygon is represented by a closed resqpy Polyline, pass Polyline.coordinates as poly
    """

    winding = 0

    for edge in range(len(poly)):
        v1 = poly[edge - 1]
        v2 = poly[edge]
        if v1[0] <= p[0] and v2[0] <= p[0]:
            continue
        if v1[1] <= p[1] and v2[1] > p[1]:
            if v1[0] > p[0] and v2[0] > p[0]:
                winding += 1
            elif p[0] < v1[0] + (v2[0] - v1[0]) * (p[1] - v1[1]) / (v2[1] - v1[1]):
                winding += 1
        elif v1[1] > p[1] and v2[1] <= p[1]:
            if v1[0] > p[0] and v2[0] > p[0]:
                winding -= 1
            elif p[0] < v1[0] + (v2[0] - v1[0]) * (p[1] - v1[1]) / (v2[1] - v1[1]):
                winding -= 1

    return bool(winding)


def pip_array_cn(p_a, poly):
    """Return bool array of 2D points inclusion, True where point is inside polygon.
    
    Uses crossing number algorithm.

    arguments:
       p_a (2D numpy float array): set of xy points to test for inclusion in polygon
       poly (2D numpy array, tuple or list of tuples or lists of at least 2 floats): the xy points
          defining a polygon

    returns:
       numpy boolean vector: True where corresponding point in p_a is within the polygon

    note:
       if the polygon is represented by a closed resqpy Polyline, pass Polyline.coordinates as poly
    """

    # p_array should be a numpy array of 2 or more axes; the final axis has extent at least 2, being x, y, ...
    # returned boolean array has shape of p_array less the final axis

    elements = np.prod(list(p_a.shape)[:-1], dtype = int)
    p = p_a.reshape((elements, -1))
    crossings = np.zeros((elements,), dtype = int)

    np_err_dict = np.geterr()
    np.seterr(divide = 'ignore', invalid = 'ignore')
    for edge in range(len(poly)):
        v1 = poly[edge - 1]
        v2 = poly[edge]
        crossings += np.where(
            np.logical_and(
                np.logical_and(
                    np.logical_or(v1[0] > p[:, 0], v2[0] > p[:, 0]),
                    np.logical_or(np.logical_and(v1[1] <= p[:, 1], v2[1] > p[:, 1]),
                                  np.logical_and(v1[1] > p[:, 1], v2[1] <= p[:, 1]))),
                np.logical_or(np.logical_and(v1[0] > p[:, 0], v2[0] > p[:, 0]),
                              (p[:, 0] < (v1[0] + (v2[0] - v1[0]) * (p[:, 1] - v1[1]) / (v2[1] - v1[1]))))), 1, 0)
    if 'divide' in np_err_dict:
        np.seterr(divide = np_err_dict['divide'])
    if 'invalid' in np_err_dict:
        np.seterr(invalid = np_err_dict['invalid'])

    return np.array(np.bitwise_and(crossings, 1), dtype = bool).reshape(list(p_a.shape)[:-1])


def points_in_polygon(x, y, polygon_file, poly_unit_multiplier = None):
    """Takes a pair of numpy arrays x, y defining points to be tested against polygon specified in polygon_file."""

    assert x.shape == y.shape, 'x and y arrays have differing shapes or are not numpy arrays'
    assert polygon_file and os.path.exists(polygon_file), 'polygon file is missing'
    if poly_unit_multiplier is not None:
        assert poly_unit_multiplier != 0.0, 'zero multiplier for polygon units not allowed'

    try:
        polygon_list = sl.read_lines(polygon_file)
        assert len(polygon_list) > 0, 'unable to read polygon from file ' + polygon_file
        assert len(polygon_list) == 1, 'more than one polygon in file ' + polygon_file
        polygon = polygon_list[0]
        if poly_unit_multiplier is not None:
            polygon[:] *= poly_unit_multiplier
        points = np.stack((x, y), axis = -1)
        return pip_array_cn(points, polygon)
    except Exception:
        log.exception('failed to determine points in polygon')

    return None


def scan(origin, ncol, nrow, dx, dy, poly):
    """Scans lines of pixels returning 2D boolean array of points within convex polygon."""

    inside = np.zeros((nrow, ncol), dtype = bool)
    ix = np.empty((2))

    for row in range(nrow):
        ix[:] = 0.0
        y = origin[1] + dy * row
        ic = 0
        for edge in range(len(poly)):
            v1 = poly[edge - 1]
            v2 = poly[edge]
            if (v1[1] > y and v2[1] > y) or (v1[1] <= y and v2[1] <= y):
                continue
            ix[ic] = v1[0] + (v2[0] - v1[0]) * (y - v1[1]) / (v2[1] - v1[1])
            ic += 1
            if ic == 2:
                break
        if ic < 2:
            continue
        if ix[1] < ix[0]:
            sx, ex = ix[1], ix[0]
        else:
            sx, ex = ix[0], ix[1]
        scol = maths.ceil((sx - origin[0]) / dx)
        ecol = int((ex - origin[0]) / dx)
        inside[row, scol:ecol + 1] = True

    return inside
